/*************************************************************************************/
/*  ab64if.c
 *
 *  Copyright 2006 Harmonic Functions, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY HARMONIC FUNCTIONS INC. "AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL HARMONIC FUNCTIONS INC. BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *  Created by Tim Bartoo on 06/01/30.
 */
/*************************************************************************************
Note that if running on x86 processor, this code assumes
	__LITTLE_ENDIAN__ is defined.
If not, add -D__LITTLE_ENDIAN__ to the compiler flags
/*************************************************************************************/
// Includes

//#define _FKN_WINDOWS 1

#ifdef _FKN_WINDOWS
	//for windows
	#include <winsock2.h>
#else
	//for Mac
	#include <netinet/in.h>
#endif

#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <math.h>
#include "ab64if.h"

#define	AF_INET		2		/* internetwork: UDP, TCP, etc. */
#define	SOCK_DGRAM	2		/* datagram socket */
#define	SOL_SOCKET	0xffff		/* options for socket level */
#define	SO_BROADCAST	0x0020		/* permit sending of broadcast msgs */
#define SO_RCVBUF	0x1002		/* receive buffer size */

/*************************************************************************************/
// Defines


//
#define true	1
#define false	0
#define transferAckTimeout 500
#define packetBufSize 1600
#define ENadrsSize sizeof(struct sockaddr_in)
#define ENadrs (struct sockaddr *) &ABsockaddr
#define PI 3.14159265358979323846
#define TWOPI (2.0 * PI)
#define DSP1 0x7FFFFF
#define MSCHdr 0xF0, 0x7F, 0x7F, 0x02, 0x10
#define plbkHdr 0xF0, 0x7F, 0x7F, 0x02, 0x13
#define plbkMultiHdr 0xF0, 0x7F, 0x7F, 0x02, 0x11
#define sysexHdr 0xF0, 0x00, 0x00, 0x40, 0x02, 0x7F, 0x7F
#define ABBDay 864961200
#define ABDS 0x41424453
#define AFNS 0x41464E53
#define EXPL 0x4558504C
#define DIFF 0x44494646
#define DFUS 0x44465553

/*************************************************************************************/
// Globals
//
int dataSock;
struct in_addr ABaddr;
struct sockaddr_in initSockaddr;
struct sockaddr_in ABsockaddr;
unsigned char *sendPacketBufPtr;
unsigned char sendPacketBuf [packetBufSize];
unsigned char *recvPacketBufPtr;
unsigned char recvPacketBuf [packetBufSize];
long setAutoUpdateEN[2];
long endAutoUpdateEN[2];
long initAutoUpdateEN[2];
long requestInquiryEN;
long uploadAudioFileEN[4];
long requestDiskInfo;
long *autoUpdateCmd;
char IPaddressString[20];
char astr[128];
char sPath[200];
char fName[200];
struct ab64Inquiry findAB64Inquiry;
struct ab64LibError libError;
struct ab64TransferProgress transferProgress;
int recvByteCount = 0;
int initialized = false;
int errFlag;
struct timeval timeout = {0, 0};
int fileTransferActive = false;
double totalTransferSize;
double bytesTransferred;
double bytesReported;
int reportTransferProgress;
int fileTransferTimer;
long fNumber;
long fType;
long mTime;
FILE *fp;
int transferringSegment;
int packetNumber;
int totalSegments;
int channelCount;
unsigned char *tfBuf;	//file transfer buffer
unsigned char *tfBufPtr;	//file transfer buffer pointer
int packetsInBuffer;
int pktNumber;
//int fwWrite;
int lastAck;
struct ab64MessageHeader fwCmpl;
struct ab64MessageHeader fromCmpl;
int fwCmplFlag;
int fromCmplFlag;
int midiChannel = 0;
long txMIDI = 0;
long endTick;
int testDrop = 0;
int sendImmediate = 0;
unsigned char *cptr;
unsigned char setCoeffCmd[39] = {sysexHdr, 0x02, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	0, 0, 0, 0, 0, 0, 0, 0,	0xF7};
unsigned char sysex9[9] = {sysexHdr, 0, 0xF7};
unsigned char sysex10[10] = {sysexHdr, 0, 0, 0xF7};
unsigned char sysex11[11] = {sysexHdr, 0, 0, 0, 0xF7};
unsigned char sysex12[12] = {sysexHdr, 0, 0, 0, 0, 0xF7};
unsigned char sysex13[13] = {sysexHdr, 0, 0, 0, 0, 0, 0xF7};
unsigned char sysex19[19] = {sysexHdr, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xF7};
unsigned char sysex21[21] = {sysexHdr, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xF7};
unsigned char sysex57[57] = {sysexHdr, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xF7};
unsigned char MSC11[11] = {MSCHdr, 0, 0, 0, 0, 0, 0xF7};
unsigned char MSC16[16] = {MSCHdr, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xF7};
unsigned char MSCStr[64] = {MSCHdr};
unsigned char playbackStr[64] = {plbkHdr};
unsigned char playbackMultiStr[64] = {plbkMultiHdr};
unsigned char nullGoStr[7] = {plbkHdr, 0x01, 0xf7};
unsigned char matrixOffStr[11] = {MSCHdr, 0x06, 0x00, 0x16, 0x00, 0x00, 0xF7};
unsigned char matrixDiagonalStr[11] = {MSCHdr, 0x06, 0x00, 0x16, 0x01, 0x00, 0xF7};
unsigned char matrixAllOnStr[11] = {MSCHdr, 0x06, 0x00, 0x16, 0x05, 0x00, 0xF7};
unsigned char clearDrive[12] = {sysexHdr, 0x0C, 0x65, 0x5A, 0x04, 0xF7};
unsigned char clearMsgCounts[12] = {sysexHdr, 0x0C, 0x11, 0x63, 0x29, 0xF7};
unsigned char saveSettings[11] = {sysexHdr, 0x3E, 0x59, 0x00, 0xF7};
unsigned char MSCheader[5] = {MSCHdr};
unsigned char sysexPrefix[5] = {0xF0, 0x00, 0x00, 0x40, 0x02};

struct setupStoreResp {
	struct ab64MessageHeader hdr;
	long response;
	} *ssr;
struct transferAck {
	struct ab64MessageHeader hdr;
	long packetNumber;
	long response;
	} *trfAck;
typedef struct getFile {
	struct ab64MessageHeader hdr;
	long fileNumber;
	long fileType;
	long segment;
	long returncode;
	unsigned char fileBytes[1024];
	} *getFilePtr;
const double MAX_RIPPLE_DB = 0.5;
const double kBW_EPSILON_HZ = 1.0;		// Interpret 0% Bandwidth as 1 Hz.
const double kPHASE_EPSILON = 1e-7;

typedef struct scomplex {
	double real;
	double imag;
} Complex;

/*************************************************************************************/
// internal function prototypes - for internal use only
// (prototypes for functions that are externally callable are in ab64if.h)
//
void logEvent (char *evstring);
void logError (char *evstring);
void millisleep (long ms);
void autoUpdateInit (void);
void autoUpdateOn (void);
void autoUpdateOff (void);
void makeLittleEndian (ab64MsgPtr mp);
void diskInfoEndian (diskInfoPtr diPtr);
void statsEndian (statsPtr sPtr);
void showEndian (showDataPtr shPtr);
void hostToFilePath (abpathv4Ptr pPtr);
void v4fileToHostPath (abpathv4Ptr pPtr);
void hostToFileDIFF (DIFFchunkHdrPtr dcPtr);
void fileToHostDIFF (DIFFchunkHdrPtr dcPtr);
void hostToFileDFUS (DFUSchunkHdrPtr dcPtr);
void fileToHostDFUS (DFUSchunkHdrPtr dcPtr);
void fileToHostChunks (int optChunks, int endSho);
void hostToFileChunks (int optChunks, int endSho);
void v5hostToFileCue (abcuev5Ptr cPtr);
void v5fileToHostCue (abcuev5Ptr cPtr);
void sendSetupStore (long fileNumber, long fileType, long segments);
void setupStoreReceived (int response);
void nextSegmentToAB64 (long ack);
void nextSegmentFromAB64 (getFilePtr gfPtr);
void nextFirmwarePacketToAB64 (long ack, long packetNumber);

void locationString (long, long, long, long, long);
void cueString (long, long, long, long, long);
long setEQ (long, long, long, long, long, long);
void setPBLKLocation (long cmdCode, long plbkch, long sel, long block);
void submasterAssignInput (long input, long submaster);
void submasterAssignOutput (long output, long submaster);
long translateChannel (long inch);
long DesignFilter (long, double, double, double, long *, long *, long *, long *, long *);
char bcd (long);
Complex cmultiply (Complex, Complex);
Complex cdiv (Complex, Complex);
Complex cadd (Complex, Complex);
double cABS (Complex);

/*************************************************************************************/
long ab64LibInit (void) {

	//get a socket
	dataSock = socket (AF_INET, SOCK_DGRAM, 0);
	if (dataSock == -1) {
		sprintf (astr, "Error getting initialization socket %d", errno);
		logError (astr);
		return false;
		}

	//set broadcast option
	char optionValue = true;
	long status = setsockopt (dataSock, SOL_SOCKET, SO_BROADCAST, &optionValue, sizeof (int));
	if (status == -1) {
		sprintf (astr, "Error setting broadcast option on initialization socket %d", errno);
		logError (astr);
		return false;
		}

	//initialize constant fields in address structure used by dataSock
	ABsockaddr.sin_port = htons (0xD758);	//set port
	ABsockaddr.sin_family = AF_INET;

	//set socket buffer size - AudioBox assumes host has a 64K incoming packet buffer
	long bufsize = 0x10000;
	status = setsockopt (dataSock, SOL_SOCKET, SO_RCVBUF, (const char*)&bufsize, sizeof (long));		
	if (status == -1) {
		sprintf (astr, "Error setting buffer size option on data socket %d", errno);
		logError (astr);
		return false;
		}

	//init packet pointer
	sendPacketBufPtr = sendPacketBuf;

	libError.code = rcAB64IFerror;
	libError.length = sizeof (libError);
	fwCmpl.code = rcFWTransferCmpl;
	fwCmpl.length = 4;
	fromCmpl.code = rcFromTransferCmpl;
	fromCmpl.length = 4;

	transferProgress.code = rcTransferProgress;
	transferProgress.length = sizeof (transferProgress);

	initAutoUpdateEN[0] = htonl (0x85000008);
	initAutoUpdateEN[1] = htonl (0x8007FFFF);
	setAutoUpdateEN[0] = htonl (0x85000008);
	setAutoUpdateEN[1] = htonl (0x0007FFFF);
	endAutoUpdateEN[0] = htonl (0x85000008);
	endAutoUpdateEN[1] = 0;
	requestInquiryEN = htonl (0x6EA70004);
	requestDiskInfo = htonl (0x65000004);

	initialized = true;

	return true;
}
/*************************************************************************************/
// broadcasts a message to all AudioBoxes on the LAN
int findAllAB64s (void) {

	//first call after startup, call initialize
	if (initialized == false)
		if (ab64LibInit () == false)
			return false;

	//stop auto updates from currently connected AB
	sendto (dataSock, (const char*)endAutoUpdateEN, 8, 0, ENadrs, ENadrsSize); //disable auto update
	
	//broadcast a message to find all AudioBoxes on LAN
	struct in_addr ABadrs;
	ABadrs.s_addr = INADDR_BROADCAST;	//set IP address - broadcast
	initSockaddr.sin_addr = ABadrs;
	initSockaddr.sin_port = htons (0xD758);	//set port
	initSockaddr.sin_family = AF_INET;
	sendto (dataSock, (const char*)&requestInquiryEN, 4, 0, (struct sockaddr *) &initSockaddr, ENadrsSize);

	//wait for messages to be returned
	millisleep (500);
	return true;
}
/*************************************************************************************/
// called in a loop immediately after findAllAB64s, returns inquiry structures
// from all AB64s on LAN, or NULL if no more
struct ab64Inquiry *nextAB64Reply (void) {
	//process received messages

	int status;
	fd_set readfds;
	while (1) {
		FD_ZERO (&readfds);
		FD_SET ((unsigned int)dataSock, &readfds);
		int dataReady = select (dataSock+1, &readfds, NULL, NULL, &timeout);
		if (dataReady) {
			status = recvfrom (dataSock, (char*) &findAB64Inquiry, sizeof (struct ab64Inquiry), 0, NULL, 0);
			if (findAB64Inquiry.code == 0x6E && findAB64Inquiry.code2 == 0xA7) {	//inquiry received
				findAB64Inquiry.IPaddress = ntohl (findAB64Inquiry.IPaddress);
				findAB64Inquiry.inchannels = ntohs (findAB64Inquiry.inchannels);
				findAB64Inquiry.outchannels = ntohs (findAB64Inquiry.outchannels);
				return &findAB64Inquiry;
				}
			}
		else
			return NULL;
		}	//end while

}
/*************************************************************************************/
//connects to AB64 at specified IP address
//if 0 is passed in as an address, disconnects from AB64
void connectAB64 (long IPaddress, long updateFlags) {

	//stop auto updates from currently connected AB
	sendto (dataSock, (const char*)endAutoUpdateEN, 8, 0, ENadrs, ENadrsSize); //disable auto update

	ABaddr.s_addr = htonl (IPaddress);
	ABsockaddr.sin_addr = ABaddr;		//set IP

	if (IPaddress) {
		setAutoUpdateEN[1] = htonl (updateFlags);
		autoUpdateInit ();	//request all data, start auto-updates
		}
}
//*************************************************************************************
void allAB64Data (void) {
	sendto (dataSock, (const char*)initAutoUpdateEN, 8, 0, ENadrs, ENadrsSize); //request all data
}
//*************************************************************************************
void autoUpdateInit (void) {
	autoUpdateCmd = setAutoUpdateEN;
	sendto (dataSock, (const char*)initAutoUpdateEN, 8, 0, ENadrs, ENadrsSize); //request all data
}
//*************************************************************************************
void autoUpdateOn (void) {
	autoUpdateCmd = setAutoUpdateEN;
	sendto (dataSock, (const char*)autoUpdateCmd, 8, 0, ENadrs, ENadrsSize);
}
//*************************************************************************************
void autoUpdateOff (void) {
	autoUpdateCmd = endAutoUpdateEN;
	sendto (dataSock, (const char*)autoUpdateCmd, 8, 0, ENadrs, ENadrsSize);
}
/*************************************************************************************/
//Called by library user at least once per frame to send and receive next AB64 messages
//via ethernet. If a message is returned, nextAB64Message should be called again
//immediately until a NULL is returned.
ab64MsgPtr nextAB64Message (void) {

//	logEvent ("nextAB64Message called");

	//unless outgoing packet is empty, send it
	int bytecount = sendPacketBufPtr - sendPacketBuf;
	if (bytecount) {
		sendto (dataSock, (const char*)sendPacketBuf, bytecount, 0, ENadrs, ENadrsSize);
		sendPacketBufPtr = sendPacketBuf;
		}

	if (bytesTransferred) {	//file transfer active
		fileTransferTimer--;
		if (fileTransferTimer == 0) {	//transfer timeout
			fileTransferTimer  = transferAckTimeout;
			sendto (dataSock, (const char*)tfBuf, 1032, 0, ENadrs, ENadrsSize);	//send a packet to force NAK
			}
		}

	//if an error occurred, report error
	if (errFlag) {
		errFlag = false;
		return (ab64MsgPtr)&libError;
		}
	//report end of firmware transfer
	if (fwCmplFlag) {
		fwCmplFlag = false;
		return &fwCmpl;
		}
	//report end of file transfer from AB64
	if (fromCmplFlag) {
		fromCmplFlag = false;
		return &fromCmpl;
		}

	//get next message
	while (1) {
		if (recvByteCount == 0) {
			//not processing a packet, attempt to get next packet from socket
			recvPacketBufPtr = recvPacketBuf;
			fd_set readfds;
			FD_ZERO (&readfds);
			FD_SET ((unsigned int)dataSock, &readfds);  
			int dataReady = select (dataSock+1, &readfds, NULL, NULL, &timeout);
			if (dataReady) {
				recvByteCount = recvfrom (dataSock, (char*)recvPacketBuf, packetBufSize, 0, NULL, 0);
				}
			else {	//no network data to return
//					logEvent ("nextAB64Message, no network data");
				recvByteCount = 0;
				sendto (dataSock, (const char*)autoUpdateCmd, 8, 0, ENadrs, ENadrsSize);
				if (bytesTransferred > bytesReported && reportTransferProgress) {	//file transfer in progress
					bytesReported = bytesTransferred;
					transferProgress.progress = bytesTransferred / totalTransferSize;
					reportTransferProgress = false;
					return (ab64MsgPtr)&transferProgress;
					}
	//			logEvent ("send auto update");
				reportTransferProgress = true;
				return NULL;
				}
			}

		ab64MsgPtr mp = (ab64MsgPtr)recvPacketBufPtr;

		mp->length = ntohs (mp->length);
		if (mp->length == 0) {	 //bad packet
			sprintf (astr, "bad packet, size 0, code %X", mp->code);
			recvByteCount = 0;
			logError (astr);
			break;
			}
		if (mp->length > packetBufSize) {	 //bad packet
			sprintf (astr, "bad packet, too large, code %X", mp->code);
			recvByteCount = 0;
			logError (astr);
			break;
			}
		//advance packet pointer, decrement unprocessed byte count
		recvPacketBufPtr += mp->length;
		recvByteCount -= mp->length;
		if (recvByteCount < 0 ) {	 //bad packet
			sprintf (astr, "bad packet, byte count went negative, code %X, size %d",
				mp->code, mp->length);
			logError (astr);
			recvByteCount = 0;
			break;
			}
//		sprintf (astr, "nextAB64Message %0X size %d",	mp->code, mp->length);
//		logEvent (astr);
		int x;
		switch (mp->code) {
		case 0x6A:		//ERASE ACK
			break;
		case 0x6B:		//FIRMWARE ACK
//			trfAck = (struct transferAck *)mp;
//			nextFirmwarePacketToAB64 (ntohl (trfAck->response),
//				ntohl (trfAck->packetNumber));
			break;
		case 0x6D:		//SETUP STORE response
			ssr = (struct setupStoreResp *)mp;
			setupStoreReceived (ntohl (ssr->response));
			break;
		case 0x84:		//SEND FILE ACK
			trfAck = (struct transferAck *)mp;
//			if (fileTransferType == audioTransferToAB64)
			nextSegmentToAB64 (ntohl (trfAck->response));
			break;
		case 0x87:		//GET FILE REPLY
			nextSegmentFromAB64 ((getFilePtr) mp);
			break;
		case 0xF0:
			trfAck = 0;
			break;
		default:	//all other messages: byte swap if necessary and pass to library user
			#if defined(__LITTLE_ENDIAN__)
			makeLittleEndian (mp);
			#endif
			if (mp->code == rcFileData)	//FILE DATA, fix up dates with ABBDay offset
				for (x = 0; x <= 3; x++) {
					((getFileDataPtr)mp)->fd[x].modDate += ABBDay;
					((getFileDataPtr)mp)->fd[x].dlDate += ABBDay;
					}
			return mp;	//return pointer to data
		}	//end switch
	}	//end while (1)
	return NULL;
}
/*************************************************************************************/
//byte swap big-endian integers in AB64 structures to native little-endian form
//messages not in the switch do not need byte swapping: rcIOLevels, rcVU, rcMuteSolo,
//rcAllDataComplete, rcLCD, rcTooManyHosts
void makeLittleEndian (ab64MsgPtr mp) {
	int x;
	short *sptr;
	long *lptr;
	switch (mp->code) {
	case rcDiskInfo:
		diskInfoEndian (&((diskInfoMsgPtr)mp)->diskinfo[0]);
		diskInfoEndian (&((diskInfoMsgPtr)mp)->diskinfo[1]);
		break;
	case rcSelChannels:
		((selChanPtr)mp)->segment = ntohl (((selChanPtr)mp)->segment);
		break;
	case rcCobranet:
		sptr = &((cobranetPtr)mp)->CN1RXBundle[0];
		x = 16;
		while (x--) {
			*sptr = ntohs (*sptr);
			sptr++;
			}
		break;
	case rcIODelay:
		lptr = &((IODelayPtr)mp)->inputDelay[0];
		x = 128;
		while (x--) {
			*lptr = ntohl (*lptr);
			lptr++;
			}
		break;
	case rcPlbkPoints:
		lptr = &((plbkPointsPtr)mp)->plbkrecord[0].stopSelection;
		x = 64*4;
		while (x--) {
			*lptr = ntohl (*lptr);
			lptr++;
			}
		break;
	case rcInquiry:
		((inquiryPtr)mp)->IPaddress = ntohl (((inquiryPtr)mp)->IPaddress);
		((inquiryPtr)mp)->inchannels = ntohs (((inquiryPtr)mp)->inchannels);
		((inquiryPtr)mp)->outchannels = ntohs (((inquiryPtr)mp)->outchannels);
		break;
	case rcStats:
		statsEndian ((statsPtr)mp);
		break;
	case rcXptLevels:
		((xptLevelsPtr)mp)->startOutChannel = ntohl (((xptLevelsPtr)mp)->startOutChannel);
		break;
	case rcXptDelay:
		((xptDelayPtr)mp)->startOutChannel = ntohl (((xptDelayPtr)mp)->startOutChannel);
		lptr = &((xptDelayPtr)mp)->xptDelay[0][0];
		x = 4*64;
		while (x--) {
			*lptr = ntohl (*lptr);
			lptr++;
			}
		break;
	case rcPlayback:
		for (x= 0; x < 64; x++) {
			((plbkPtr)mp)->pch[x].selectionNumber = ntohl (((plbkPtr)mp)->pch[x].selectionNumber);
			((plbkPtr)mp)->pch[x].playbackLocation = ntohl (((plbkPtr)mp)->pch[x].playbackLocation);
			((plbkPtr)mp)->pch[x].emptyBlocks = ntohl (((plbkPtr)mp)->pch[x].emptyBlocks);
			}
		break;
	case rcShow:
		showEndian ((showDataPtr) mp);
		break;
	case rcSubMaster:
		((submasterPtr)mp)->submasterGain = ntohs (((submasterPtr)mp)->submasterGain);
		((submasterPtr)mp)->submasterZero = ntohs (((submasterPtr)mp)->submasterZero);
		break;
	case rcDiskState:
		((diskStatePtr)mp)->diskCapacity = ntohl (((diskStatePtr)mp)->diskCapacity);
		((diskStatePtr)mp)->nowfreeBlks = ntohl (((diskStatePtr)mp)->nowfreeBlks);
		((diskStatePtr)mp)->freeBlocks = ntohl (((diskStatePtr)mp)->freeBlocks);
		((diskStatePtr)mp)->highestSelNum = ntohl (((diskStatePtr)mp)->highestSelNum);
		((diskStatePtr)mp)->defragMarker = ntohl (((diskStatePtr)mp)->defragMarker);
		((diskStatePtr)mp)->defragEnd = ntohl (((diskStatePtr)mp)->defragEnd);
		((diskStatePtr)mp)->highestShNum = ntohl (((diskStatePtr)mp)->highestShNum);
		((diskStatePtr)mp)->MIDIgroupIDs = ntohs (((diskStatePtr)mp)->MIDIgroupIDs);
		((diskStatePtr)mp)->GPItrigLockout = ntohl (((diskStatePtr)mp)->GPItrigLockout);
		break;
	case rcEQ:
		((eqPtr)mp)->IOflag = ntohl (((eqPtr)mp)->IOflag);
		((eqPtr)mp)->channel = ntohl (((eqPtr)mp)->channel);
		((eqPtr)mp)->eqch.nSections = ntohl (((eqPtr)mp)->eqch.nSections);
		for (x = 0; x < 7; x++) {
			((eqPtr)mp)->eqch.section[x].a2 = ntohl (((eqPtr)mp)->eqch.section[x].a2);
			((eqPtr)mp)->eqch.section[x].a1 = ntohl (((eqPtr)mp)->eqch.section[x].a1);
			((eqPtr)mp)->eqch.section[x].b1 = ntohl (((eqPtr)mp)->eqch.section[x].b1);
			((eqPtr)mp)->eqch.section[x].a0 = ntohl (((eqPtr)mp)->eqch.section[x].a0);
			((eqPtr)mp)->eqch.section[x].b0 = ntohl (((eqPtr)mp)->eqch.section[x].b0);
			((eqPtr)mp)->eqch.section[x].filterType = ntohs (((eqPtr)mp)->eqch.section[x].filterType);
			((eqPtr)mp)->eqch.section[x].bandwidth = ntohs (((eqPtr)mp)->eqch.section[x].bandwidth);
			((eqPtr)mp)->eqch.section[x].gain = ntohs (((eqPtr)mp)->eqch.section[x].gain);
			((eqPtr)mp)->eqch.section[x].frequency = ntohl (((eqPtr)mp)->eqch.section[x].frequency);
			}
		break;
	case rcDirectory:
		((directoryPtr)mp)->dirType = ntohl (((directoryPtr)mp)->dirType);
		((directoryPtr)mp)->directorySegment = ntohl (((directoryPtr)mp)->directorySegment);
		lptr = &((directoryPtr)mp)->blocks[0];
		x = 128;
		while (x--) {
			*lptr = ntohl (*lptr);
			lptr++;
			}
		break;
	case rcDirUpdateCmpl:
		((dirUpdateCmplPtr)mp)->dirType = ntohl (((dirUpdateCmplPtr)mp)->dirType);
		((dirUpdateCmplPtr)mp)->fileNumber = ntohl (((dirUpdateCmplPtr)mp)->fileNumber);
		break;
	case rcRealTime:
		((realTimePtr)mp)->validFlags = ntohs (((realTimePtr)mp)->validFlags);
		((realTimePtr)mp)->sunrise = ntohs (((realTimePtr)mp)->sunrise);
		((realTimePtr)mp)->sunset = ntohs (((realTimePtr)mp)->sunset);
		break;
	case rcFileData:
		((getFileDataPtr)mp)->ftype = ntohl (((getFileDataPtr)mp)->ftype);
		((getFileDataPtr)mp)->stfn = ntohl (((getFileDataPtr)mp)->stfn);
		for (x = 0; x <= 3; x++) {
			((getFileDataPtr)mp)->fd[x].dlDate = ntohl (((getFileDataPtr)mp)->fd[x].dlDate);
			((getFileDataPtr)mp)->fd[x].modDate = ntohl (((getFileDataPtr)mp)->fd[x].modDate);
			((getFileDataPtr)mp)->fd[x].sourceFileSize = ntohl (((getFileDataPtr)mp)->fd[x].sourceFileSize);
			}
		break;
	}	//end switch
}
/*************************************************************************************/
//swap byte order: incoming disk info structure
void diskInfoEndian (diskInfoPtr diPtr) {
	diPtr->errorNumber = ntohs (diPtr->errorNumber);
	diPtr->resetDiagnosticCode = ntohs (diPtr->resetDiagnosticCode);
	diPtr->capacityHi = ntohs (diPtr->capacityHi);
	diPtr->capacityLo = ntohs (diPtr->capacityLo);
	diPtr->cylinders = ntohs (diPtr->cylinders);
	diPtr->specificConfig = ntohs (diPtr->specificConfig);
	diPtr->heads = ntohs (diPtr->heads);
	diPtr->sectorsPerTrack = ntohs (diPtr->sectorsPerTrack);
	diPtr->sectorsPerMultCmd = ntohs (diPtr->sectorsPerMultCmd);
	diPtr->capabilities1 = ntohs (diPtr->capabilities1);
	diPtr->capabilities2 = ntohs (diPtr->capabilities2);
	diPtr->validBits = ntohs (diPtr->validBits);
	diPtr->currentCylinders = ntohs (diPtr->currentCylinders);
	diPtr->currentHeads = ntohs (diPtr->currentHeads);
	diPtr->currentSectorsPerTrack = ntohs (diPtr->currentSectorsPerTrack);
	diPtr->capacityLowWord = ntohs (diPtr->capacityLowWord);
	diPtr->capacityHighWord = ntohs (diPtr->capacityHighWord);
	diPtr->sectorsPerInterrupt = ntohs (diPtr->sectorsPerInterrupt);
	diPtr->addressableSectorsLowWord = ntohs (diPtr->addressableSectorsLowWord);
	diPtr->addressableSectorsHighWord = ntohs (diPtr->addressableSectorsHighWord);
	diPtr->DMAmode = ntohs (diPtr->DMAmode);
	diPtr->PIOModes = ntohs (diPtr->PIOModes);
	diPtr->minDMAcycle = ntohs (diPtr->minDMAcycle);
	diPtr->recommendDMAcycle = ntohs (diPtr->recommendDMAcycle);
	diPtr->minPIOwithoutFC = ntohs (diPtr->minPIOwithoutFC);
	diPtr->minPIOwithFC = ntohs (diPtr->minPIOwithFC);
	diPtr->queueDepth = ntohs (diPtr->queueDepth);
	diPtr->majorVersion = ntohs (diPtr->majorVersion);
	diPtr->minorVersion = ntohs (diPtr->minorVersion);
	diPtr->cmdSetSupportedBits1 = ntohs (diPtr->cmdSetSupportedBits1);
	diPtr->cmdSetSupportedBits2 = ntohs (diPtr->cmdSetSupportedBits2);
	diPtr->cmdSetSupportedBits3 = ntohs (diPtr->cmdSetSupportedBits3);
	diPtr->cmdSetEnabledBits1 = ntohs (diPtr->cmdSetEnabledBits1);
	diPtr->cmdSetEnabledBits2 = ntohs (diPtr->cmdSetEnabledBits2);
	diPtr->cmdSetEnabledBits3 = ntohs (diPtr->cmdSetEnabledBits3);
	diPtr->ultraModes = ntohs (diPtr->ultraModes);
	diPtr->securityEraseTime = ntohs (diPtr->securityEraseTime);
	diPtr->advancedPowerManValue = ntohs (diPtr->advancedPowerManValue);
	diPtr->masterPasswordRevCode = ntohs (diPtr->masterPasswordRevCode);
	diPtr->hardwareResetResult = ntohs (diPtr->hardwareResetResult);
	diPtr->acousticManValue = ntohs (diPtr->acousticManValue);
	diPtr->maxLBA48word1 = ntohs (diPtr->maxLBA48word1);
	diPtr->maxLBA48word2 = ntohs (diPtr->maxLBA48word2);
	diPtr->maxLBA48word3 = ntohs (diPtr->maxLBA48word3);
	diPtr->maxLBA48word4 = ntohs (diPtr->maxLBA48word4);
}
/*************************************************************************************/
//set byte order: incoming stats structure
void statsEndian (statsPtr sPtr) {
	sPtr->freeBuffers = ntohl (sPtr->freeBuffers);
	sPtr->lowWaterMark = ntohl (sPtr->lowWaterMark);
	sPtr->timerCount = ntohl (sPtr->timerCount);
	sPtr->workCount = ntohl (sPtr->workCount);
	sPtr->MIDI0TXCount = ntohl (sPtr->MIDI0TXCount);
	sPtr->MIDI1TXCount = ntohl (sPtr->MIDI1TXCount);
	sPtr->currentTime = ntohl (sPtr->currentTime);
	sPtr->cpuUtilization = ntohl (sPtr->cpuUtilization);
	sPtr->enetMsgs = ntohl (sPtr->enetMsgs);
	sPtr->enetOKmsgs = ntohl (sPtr->enetOKmsgs);
	sPtr->enetBADmsgs = ntohl (sPtr->enetBADmsgs);
	sPtr->MIDI0rxMsgs = ntohl (sPtr->MIDI0rxMsgs);
	sPtr->MIDI0OKmsgs = ntohl (sPtr->MIDI0OKmsgs);
	sPtr->MIDI0BADmsgs = ntohl (sPtr->MIDI0BADmsgs);
	sPtr->MIDI1rxMsgs = ntohl (sPtr->MIDI1rxMsgs);
	sPtr->MIDI1OKmsgs = ntohl (sPtr->MIDI1OKmsgs);
	sPtr->MIDI1BADmsgs = ntohl (sPtr->MIDI1BADmsgs);
	sPtr->MIDI0txMsgs = ntohl (sPtr->MIDI0txMsgs);
	sPtr->MIDI1txMsgs = ntohl (sPtr->MIDI1txMsgs);
	sPtr->temperature = ntohl (sPtr->temperature);
}
/*************************************************************************************/
//set byte order: incoming show structure
void showEndian (showDataPtr shPtr) {
	shPtr->showNumber = ntohs (shPtr->showNumber);
	shPtr->numPaths = ntohs (shPtr->numPaths);
	shPtr->numLists = ntohs (shPtr->numLists);
	shPtr->MTCgenerator = ntohl (shPtr->MTCgenerator);
	int x = runlistsAllocated;
	struct ablistdata *listPtr = &shPtr->listdata[0];
	while (x--) {
		if (listPtr->dataValid) {
			listPtr->listNumber = ntohs (listPtr->listNumber);
			listPtr->listFlags = ntohs (listPtr->listFlags);
			listPtr->pathNumber = ntohs (listPtr->pathNumber);
			listPtr->nCues = ntohs (listPtr->nCues);
			listPtr->nLiveCues = ntohs (listPtr->nLiveCues);
			listPtr->listClock = ntohl (listPtr->listClock);
			listPtr->seqClock = ntohl (listPtr->seqClock);
			listPtr->lastCuenumber = ntohs (listPtr->lastCuenumber);
			listPtr->nextCuenumber = ntohs (listPtr->nextCuenumber);
			}
		listPtr++;
		}
}
/*************************************************************************************/
//set byte order when saving v5 show file
//on entry, show file is in native byte order in memory at shPtr, size is shSize bytes
//on completion, show file is in storage byte order (big endian)
void hostTov5ShowFile (void *shPtr, int shSize) {
	abv5shHdrPtr shHdr = (abv5shHdrPtr)shPtr;
	hostToFileChunks ((int)shPtr + shHdr->optchunkofs, (int) shPtr + shSize);

	//byte swap path table, path index, paths, cue index and cues
	abv5pathTblPtr pathTblPtr = (abv5pathTblPtr)((long)shPtr + abshowhdrsizev5);
	pathTblPtr->size = htonl (pathTblPtr->size);
	pathTblPtr->structID = htonl (pathTblPtr->structID);
	abv5pathindexPtr pathIndxPtr = (abv5pathindexPtr)((long)pathTblPtr + abpathtablesizev5);
	int pathIndexCount = shHdr->npaths;
	abpathv5Ptr pathPtr;
	abcueindexv5Ptr cueIndexPtr;
	int cueCount;
	while (pathIndexCount--) {
		pathPtr = (abpathv5Ptr)((int)shPtr + pathIndxPtr->address);
		pathIndxPtr->listnumber = htons (pathIndxPtr->listnumber);
		pathIndxPtr->pathnumber = htons (pathIndxPtr->pathnumber);
		pathIndxPtr->address = htonl (pathIndxPtr->address);
		pathIndxPtr->size = htonl (pathIndxPtr->size);
		cueCount = pathPtr->ncues;
		pathPtr->size = htonl (pathPtr->size);
		pathPtr->structID = htonl (pathPtr->structID);
		pathPtr->listnumber = htons (pathPtr->listnumber);
		pathPtr->pathnumber = htons (pathPtr->pathnumber);
		pathPtr->ncues = htonl (pathPtr->ncues);
		pathPtr->listFlags = htonl (pathPtr->listFlags);
		pathPtr->listNoteOffMask = htonl (pathPtr->listNoteOffMask);
		cueIndexPtr = (abcueindexv5Ptr)((long)pathPtr + abpathsizev5);
		while (cueCount--) {
			v5hostToFileCue ((abcuev5Ptr)((long)shPtr + cueIndexPtr->address));
			cueIndexPtr->cuenumber = htonl (cueIndexPtr->cuenumber);
			cueIndexPtr->address = htonl (cueIndexPtr->address);
			cueIndexPtr->size = htonl (cueIndexPtr->size);
			cueIndexPtr++;
			}
		pathIndxPtr++;
		}

	//byte swap setup cue
	if (shHdr->setupCueOffset)
		v5hostToFileCue ((abcuev5Ptr)((long)shPtr + shHdr->setupCueOffset));

	//byte swap show header
	shHdr->fileID = htonl (shHdr->fileID);
	shHdr->vers = htons (shHdr->vers);
	shHdr->nlists = htonl (shHdr->nlists);
	shHdr->npaths = htonl (shHdr->npaths);
	shHdr->optchunkofs = htonl (shHdr->optchunkofs);
	shHdr->setupCueOffset = htonl (shHdr->setupCueOffset);
	shHdr->setupCueSize = htonl (shHdr->setupCueSize);
}
/*************************************************************************************/
//set byte order for path structure when opening show
void v5hostToFileCue (abcuev5Ptr cPtr) {
	cPtr->size = htonl (cPtr->size);
	cPtr->structID = htonl (cPtr->structID);
	cPtr->cuenumber = htonl (cPtr->cuenumber);
	cPtr->cuetime = htonl (cPtr->cuetime);
	cPtr->flags = htonl (cPtr->flags);
	cPtr->nevents = htonl (cPtr->nevents);
}
/*************************************************************************************/
//set byte order when saving v4 show file
//on entry, show file is in native byte order in memory at shPtr, size is shSize bytes
//on completion, show file is in storage byte order (big endian)
void hostTov4ShowFile (void *shPtr, int shSize) {
	abv4shHdrPtr v4ShowHeader = (abv4shHdrPtr)shPtr;
	hostToFileChunks ((int)shPtr + v4ShowHeader->optchunkofs, (int) shPtr + shSize);

	//byte swap paths & path index
	if (v4ShowHeader->setupPathOffset)
		hostToFilePath ((abpathv4Ptr)((long)shPtr + v4ShowHeader->setupPathOffset));

	struct abpathindexv4 *piPtr = (struct abpathindexv4 *)
		((long)shPtr + abshowhdrsizev4 + (128 *  ablistsize));
	int loopCount = v4ShowHeader->npaths;
	while (loopCount--) {
		hostToFilePath ((abpathv4Ptr)((long)shPtr + piPtr->pathoffset));
		piPtr->listnumber = htons (piPtr->listnumber);
		piPtr->pathnumber = htons (piPtr->pathnumber);
		piPtr->pathoffset = htonl (piPtr->pathoffset);
		piPtr++;
		}

	//byte swap list array
	struct ablist *lstArrPtr = (struct ablist *)((long)shPtr + abshowhdrsizev4);
	loopCount = 128;
	while (loopCount--) {
		lstArrPtr->noteOffMask = htons (lstArrPtr->noteOffMask);
		lstArrPtr->flags = htons (lstArrPtr->flags);
		lstArrPtr++;
		}

	//byte swap show header
	v4ShowHeader->fileID = htonl (v4ShowHeader->fileID);
	v4ShowHeader->vers = htons (v4ShowHeader->vers);
	v4ShowHeader->npaths = htons (v4ShowHeader->npaths);
	v4ShowHeader->optchunkofs = htonl (v4ShowHeader->optchunkofs);
	v4ShowHeader->setupPathOffset = htonl (v4ShowHeader->setupPathOffset);
}
/*************************************************************************************/
//set byte order for optional chunks when opening show
void hostToFileChunks (int optChunks, int endSho) {
	//find ABDS, AFNS, EXPL, DIFF and DFUS chunks and byte swap
	int x = optChunks;
	descHdrPtr chunkPtr;
	while (optChunks <= x && x < endSho) {
		chunkPtr = (descHdrPtr) x;
		x = x + 8 + chunkPtr->chunkSize;
		if (chunkPtr->chunkid == ABDS
			|| chunkPtr->chunkid == AFNS
			|| chunkPtr->chunkid == EXPL) {
			chunkPtr->chunkid = htonl (chunkPtr->chunkid);
			chunkPtr->chunkSize = htonl (chunkPtr->chunkSize);
			chunkPtr->listcount = htons (chunkPtr->listcount);
			chunkPtr->othercountHi = htons (chunkPtr->othercountHi);
			chunkPtr->othercountLo = htons (chunkPtr->othercountLo);
			}
		if (chunkPtr->chunkid == DIFF)
			hostToFileDIFF ((DIFFchunkHdrPtr)chunkPtr);
		if (chunkPtr->chunkid == DFUS)
			hostToFileDFUS ((DFUSchunkHdrPtr)chunkPtr);
		}
}
/*************************************************************************************/
//set byte order when saving path
void hostToFilePath (abpathv4Ptr pPtr) {
	abcuev4Ptr cPtr = (abcuev4Ptr) ((long)pPtr + sizeof (struct abpathv4));
	abeventv4Ptr ePtr = (abeventv4Ptr) ((long)pPtr + pPtr->evHeapOfs);
	int cueCount = pPtr->ncues;
	while (cueCount--) {
		int eventCount = cPtr->nevents;
		while (eventCount--) {
			long x = (long)ePtr + abeventv4HeaderSize + ePtr->evsize;
			if (x & 1) x++;
			ePtr->evsize = htons (ePtr->evsize);
			ePtr->evdelayHi = htons (ePtr->evdelayHi);
			ePtr->evdelayLo = htons (ePtr->evdelayLo);
			//move event pointer to next event
			ePtr = (abeventv4Ptr)x;
			}
		cPtr->cuetime = htonl (cPtr->cuetime);
		cPtr->event1offset = htonl (cPtr->event1offset);
		cPtr->nevents = htons (cPtr->nevents);
		cPtr->flags = htons (cPtr->flags);
		cPtr->cuenumber = htons (cPtr->cuenumber);
		cPtr = (abcuev4Ptr)((long)cPtr + abcuev4size);
		}
	pPtr->structID = htonl (pPtr->structID);
	pPtr->ncues = htons (pPtr->ncues);
	pPtr->evHeapOfs = htons (pPtr->evHeapOfs);
}
/*************************************************************************************/
//set byte order when saving DFUS chunk
void hostToFileDFUS (DFUSchunkHdrPtr dcPtr) {
	dcPtr->chunkid = htonl (dcPtr->chunkid);
	dcPtr->chunkSize = htonl (dcPtr->chunkSize);
	int spkrMapCount = dcPtr->numberOfSpkrMaps;
	int diffusionCount = dcPtr->numberOfDiffusions;
	dcPtr->numberOfSpkrMaps = htons (dcPtr->numberOfSpkrMaps);
	dcPtr->numberOfDiffusions = htons (dcPtr->numberOfDiffusions);
	DFUSspkrMapPtr smPtr = (DFUSspkrMapPtr) ((long)dcPtr + DFUSchunkHdrSize);
	while (spkrMapCount--) {
		int spkrCount = smPtr->numberOfSpkrs;
		smPtr->originX = htonl (smPtr->originX);
		smPtr->originY = htonl (smPtr->originY);
		smPtr->numberOfSpkrs = htonl (smPtr->numberOfSpkrs);
		DFUSspkrPtr sPtr = (DFUSspkrPtr) ((long)smPtr + DFUSspeakerMapSize);
		while (spkrCount--) {
			sPtr->outputCh = htonl (sPtr->outputCh);
			sPtr->x = htonl (sPtr->x);
			sPtr->y = htonl (sPtr->y);
			sPtr++;
			}
		smPtr = (DFUSspkrMapPtr) sPtr;
		}

	DFUSdiffPtr dPtr = (DFUSdiffPtr) smPtr;
	while (diffusionCount--) {
		int trajectoryCount = dPtr->numberOfTrajectories;
		dPtr->startEvent = htonl (dPtr->startEvent);
		dPtr->numberOfEvents = htonl (dPtr->numberOfEvents);
		dPtr->numberOfTrajectories = htonl (dPtr->numberOfTrajectories);
		DFUStrajPtr trPtr = (DFUStrajPtr) ((long)dPtr + DFUSdiffusionSize);
		while (trajectoryCount--) {
			int pointCount = trPtr->numberOfPoints;
			trPtr->numberOfPoints = htons (trPtr->numberOfPoints);
			trPtr->inputCh = htons (trPtr->inputCh);
			DFUSsndPntPtr spPtr = (DFUSsndPntPtr) ((long)trPtr + DFUStrajectorySize);
			while (pointCount--) {
				spPtr->x = htonl (spPtr->x);
				spPtr->y = htonl (spPtr->y);
				spPtr->diameter = htonl (spPtr->diameter);
				spPtr->transitionTime = htonl (spPtr->transitionTime);
				spPtr->pauseTime = htonl (spPtr->pauseTime);
				spPtr++;
				}
			trPtr = (DFUStrajPtr) spPtr;
			}
		dPtr = (DFUSdiffPtr) trPtr;
		}
}
/*************************************************************************************/
//set byte order when saving DIFF chunk
void hostToFileDIFF (DIFFchunkHdrPtr dcPtr) {
	dcPtr->chunkid = htonl (dcPtr->chunkid);
	dcPtr->chunkSize = htonl (dcPtr->chunkSize);
	int spkrMapCount = dcPtr->numberOfSpkrMaps;
	int trajectoryCount = dcPtr->numberOfTrajectories;
	dcPtr->numberOfSpkrMaps = htons (dcPtr->numberOfSpkrMaps);
	dcPtr->numberOfTrajectories = htons (dcPtr->numberOfTrajectories);
	DFUSspkrMapPtr smPtr = (DFUSspkrMapPtr) ((long)dcPtr + DIFFchunkHdrSize);
	while (spkrMapCount--) {
		int spkrCount = smPtr->numberOfSpkrs;
		smPtr->originX = htonl (smPtr->originX);
		smPtr->originY = htonl (smPtr->originY);
		smPtr->numberOfSpkrs = htonl (smPtr->numberOfSpkrs);
		DFUSspkrPtr sPtr = (DFUSspkrPtr) ((long)smPtr + DFUSspeakerMapSize);
		while (spkrCount--) {
			sPtr->outputCh = htonl (sPtr->outputCh);
			sPtr->x = htonl (sPtr->x);
			sPtr->y = htonl (sPtr->y);
			sPtr++;
			}
		smPtr = (DFUSspkrMapPtr) sPtr;
		}
	DIFFtrajPtr trPtr = (DIFFtrajPtr) smPtr;
	while (trajectoryCount--) {
		int pointCount = trPtr->numberOfPoints;
		trPtr->cueNumber = htonl (trPtr->cueNumber);
		trPtr->cueSubNumber = htonl (trPtr->cueSubNumber);
		trPtr->numberOfPoints = htons (trPtr->numberOfPoints);
		trPtr->inputCh = htons (trPtr->inputCh);
		trPtr->listNumber = htons (trPtr->listNumber);
		trPtr->pathNumber = htons (trPtr->pathNumber);
		DFUSsndPntPtr spPtr = (DFUSsndPntPtr) ((long)trPtr + DIFFtrajectorySize);
		while (pointCount--) {
			spPtr->x = htonl (spPtr->x);
			spPtr->y = htonl (spPtr->y);
			spPtr->diameter = htonl (spPtr->diameter);
			spPtr->transitionTime = htonl (spPtr->transitionTime);
			spPtr->pauseTime = htonl (spPtr->pauseTime);
			spPtr++;
			}
		trPtr = (DIFFtrajPtr) spPtr;
		}
}
/*************************************************************************************/
//set byte order when opening version 5 show file
//on entry, show file is in storage byte order (big endian) in memory at shPtr,
// size is shSize bytes. On completion, show file is in native byte order
void v5showFileToHost (void *shPtr, int shSize) {
	//byte swap show header
	struct abshowhdrv5 *v5ShowHeader = (struct abshowhdrv5 *)shPtr;
	v5ShowHeader->fileID = ntohl (v5ShowHeader->fileID);
	v5ShowHeader->vers = ntohs (v5ShowHeader->vers);
	v5ShowHeader->nlists = ntohl (v5ShowHeader->nlists);
	v5ShowHeader->npaths = ntohl (v5ShowHeader->npaths);
	v5ShowHeader->optchunkofs = ntohl (v5ShowHeader->optchunkofs);
	v5ShowHeader->setupCueOffset = ntohl (v5ShowHeader->setupCueOffset);
	v5ShowHeader->setupCueSize = ntohl (v5ShowHeader->setupCueSize);

	//byte swap setup cue
	if (v5ShowHeader->setupCueOffset)
		v5fileToHostCue ((abcuev5Ptr)((long)shPtr + v5ShowHeader->setupCueOffset));

	//byte swap path table, path index, paths, cue index and cues
	abv5pathTblPtr pathTblPtr = (abv5pathTblPtr)((long)shPtr + abshowhdrsizev5);
	pathTblPtr->size = ntohl (pathTblPtr->size);
	pathTblPtr->structID = ntohl (pathTblPtr->structID);
	abv5pathindexPtr pathIndxPtr = (abv5pathindexPtr)((long)pathTblPtr + abpathtablesizev5);
	int pathIndexCount = v5ShowHeader->npaths;
	abpathv5Ptr pathPtr;
	abcueindexv5Ptr cueIndexPtr;
	int cueCount;
	while (pathIndexCount--) {
		pathIndxPtr->listnumber = ntohs (pathIndxPtr->listnumber);
		pathIndxPtr->pathnumber = ntohs (pathIndxPtr->pathnumber);
		pathIndxPtr->address = ntohl (pathIndxPtr->address);
		pathIndxPtr->size = ntohl (pathIndxPtr->size);
		pathPtr = (abpathv5Ptr)((int)shPtr + pathIndxPtr->address);
		pathPtr->size = ntohl (pathPtr->size);
		pathPtr->structID = ntohl (pathPtr->structID);
		pathPtr->listnumber = ntohs (pathPtr->listnumber);
		pathPtr->pathnumber = ntohs (pathPtr->pathnumber);
		pathPtr->ncues = ntohl (pathPtr->ncues);
		pathPtr->listFlags = ntohl (pathPtr->listFlags);
		pathPtr->listNoteOffMask = ntohl (pathPtr->listNoteOffMask);
		cueIndexPtr = (abcueindexv5Ptr)((long)pathPtr + abpathsizev5);
		cueCount = pathPtr->ncues;
		while (cueCount--) {
			cueIndexPtr->cuenumber = ntohl (cueIndexPtr->cuenumber);
			cueIndexPtr->address = ntohl (cueIndexPtr->address);
			cueIndexPtr->size = ntohl (cueIndexPtr->size);
			v5fileToHostCue ((abcuev5Ptr)((long)shPtr + cueIndexPtr->address));
			cueIndexPtr++;
			}
		pathIndxPtr++;
		}
	
	//byte swap optional chunks
	fileToHostChunks ((int)shPtr + v5ShowHeader->optchunkofs, (int)shPtr + shSize);
}
/*************************************************************************************/
//set byte order for path structure when opening show
void v5fileToHostCue (abcuev5Ptr cPtr) {
	cPtr->size = ntohl (cPtr->size);
	cPtr->structID = ntohl (cPtr->structID);
	cPtr->cuenumber = ntohl (cPtr->cuenumber);
	cPtr->cuetime = ntohl (cPtr->cuetime);
	cPtr->flags = ntohl (cPtr->flags);
	cPtr->nevents = ntohl (cPtr->nevents);
}
/*************************************************************************************/
//set byte order when opening version 4 show file
//on entry, show file is in storage byte order (big endian) in memory at shPtr,
// size is shSize bytes. On completion, show file is in native byte order
void v4showFileToHost (void *shPtr, int shSize) {
	//byte swap show header
	struct abshowhdrv4 *v4ShowHeader = (struct abshowhdrv4 *)shPtr;
	v4ShowHeader->fileID = ntohl (v4ShowHeader->fileID);
	v4ShowHeader->vers = ntohs (v4ShowHeader->vers);
	v4ShowHeader->npaths = ntohs (v4ShowHeader->npaths);
	v4ShowHeader->optchunkofs = ntohl (v4ShowHeader->optchunkofs);
	v4ShowHeader->setupPathOffset = ntohl (v4ShowHeader->setupPathOffset);

	//byte swap list array
	struct ablist *lstArrPtr = (struct ablist *)((long)shPtr + abshowhdrsizev4);
	int loopCount = 128;
	while (loopCount--) {
		lstArrPtr->noteOffMask = ntohs (lstArrPtr->noteOffMask);
		lstArrPtr->flags = ntohs (lstArrPtr->flags);
		lstArrPtr++;
		}

	//byte swap paths & path index
	if (v4ShowHeader->setupPathOffset)
		v4fileToHostPath ((abpathv4Ptr)((long)shPtr + v4ShowHeader->setupPathOffset));
	struct abpathindexv4 *piPtr = (struct abpathindexv4 *)
		((long)shPtr + abshowhdrsizev4 + (128 *  ablistsize));
	loopCount = v4ShowHeader->npaths;
	while (loopCount--) {
		piPtr->listnumber = ntohs (piPtr->listnumber);
		piPtr->pathnumber = ntohs (piPtr->pathnumber);
		piPtr->pathoffset = ntohl (piPtr->pathoffset);
		v4fileToHostPath ((abpathv4Ptr)((long)shPtr + piPtr->pathoffset));
		piPtr++;
		}

	//byte swap optional chunks
	fileToHostChunks ((int)shPtr + v4ShowHeader->optchunkofs, (int) shPtr + shSize);
}
/*************************************************************************************/
//set byte order for optional chunks when opening show
void fileToHostChunks (int optChunks, int endSho) {
	//find ABDS, AFNS, EXPL, DIFF and DFUS chunks and byte swap
	descHdrPtr chunkPtr;
	int x = optChunks;
	while (optChunks <= x && x < endSho) {
		chunkPtr = (descHdrPtr) x;
		chunkPtr->chunkid = ntohl (chunkPtr->chunkid);
		chunkPtr->chunkSize = ntohl (chunkPtr->chunkSize);
		if (chunkPtr->chunkid == ABDS
			|| chunkPtr->chunkid == AFNS
			|| chunkPtr->chunkid == EXPL) {
			chunkPtr->listcount = ntohs (chunkPtr->listcount);
			chunkPtr->othercountHi = ntohs (chunkPtr->othercountHi);
			chunkPtr->othercountLo = ntohs (chunkPtr->othercountLo);
			}
		if (chunkPtr->chunkid == DIFF)
			fileToHostDIFF ((DIFFchunkHdrPtr)chunkPtr);
		if (chunkPtr->chunkid == DFUS)
			fileToHostDFUS ((DFUSchunkHdrPtr)chunkPtr);
		x = x + 8 + chunkPtr->chunkSize;
		}
}
/*************************************************************************************/
//set byte order for path structure when opening show
void v4fileToHostPath (abpathv4Ptr pPtr) {
	pPtr->structID = ntohl (pPtr->structID);
	pPtr->ncues = ntohs (pPtr->ncues);
	pPtr->evHeapOfs = ntohs (pPtr->evHeapOfs);
	abcuev4Ptr cPtr = (abcuev4Ptr) ((long)pPtr + abpathsizev4);
	abeventv4Ptr ePtr = (abeventv4Ptr) ((long)pPtr + pPtr->evHeapOfs);
	int cueCount = pPtr->ncues;
	while (cueCount--) {
		cPtr->cuetime = ntohl (cPtr->cuetime);
		cPtr->event1offset = ntohl (cPtr->event1offset);
		cPtr->nevents = ntohs (cPtr->nevents);
		cPtr->flags = ntohs (cPtr->flags);
		cPtr->cuenumber = ntohs (cPtr->cuenumber);
		int eventCount = cPtr->nevents;
		while (eventCount--) {
			ePtr->evsize = ntohs (ePtr->evsize);
			ePtr->evdelayHi = ntohs (ePtr->evdelayHi);
			ePtr->evdelayLo = ntohs (ePtr->evdelayLo);
			//move event pointer to next event
			long x = (long)ePtr + abeventv4HeaderSize + ePtr->evsize;
			if (x & 1) x++;
			ePtr = (abeventv4Ptr)x;
			}
		cPtr = (abcuev4Ptr)((unsigned)cPtr + abcuev4size);
		}
}
/*************************************************************************************/
//set byte order for DFUS chunk when opening show
void fileToHostDFUS (DFUSchunkHdrPtr dcPtr) {
	dcPtr->numberOfSpkrMaps = ntohs (dcPtr->numberOfSpkrMaps);
	int spkrMapCount = dcPtr->numberOfSpkrMaps;
	dcPtr->numberOfDiffusions = ntohs (dcPtr->numberOfDiffusions);
	int diffusionCount = dcPtr->numberOfDiffusions;

	DFUSspkrMapPtr smPtr = (DFUSspkrMapPtr) ((long)dcPtr + DFUSchunkHdrSize);
	while (spkrMapCount--) {
		smPtr->originX = ntohl (smPtr->originX);
		smPtr->originY = ntohl (smPtr->originY);
		smPtr->numberOfSpkrs = ntohl (smPtr->numberOfSpkrs);
		int spkrCount = smPtr->numberOfSpkrs;
		DFUSspkrPtr sPtr = (DFUSspkrPtr) ((long)smPtr + DFUSspeakerMapSize);
		while (spkrCount--) {
			sPtr->outputCh = ntohl (sPtr->outputCh);
			sPtr->x = ntohl (sPtr->x);
			sPtr->y = ntohl (sPtr->y);
			sPtr++;
			}
		smPtr = (DFUSspkrMapPtr) sPtr;
		}

	DFUSdiffPtr dPtr = (DFUSdiffPtr) smPtr;
	while (diffusionCount--) {
		dPtr->startEvent = ntohl (dPtr->startEvent);
		dPtr->numberOfEvents = ntohl (dPtr->numberOfEvents);
		dPtr->numberOfTrajectories = ntohl (dPtr->numberOfTrajectories);
		int trajectoryCount = dPtr->numberOfTrajectories;
		DFUStrajPtr trPtr = (DFUStrajPtr) ((long)dPtr + DFUSdiffusionSize);
		while (trajectoryCount--) {
			trPtr->inputCh = ntohs (trPtr->inputCh);
			trPtr->numberOfPoints = ntohs (trPtr->numberOfPoints);
			int pointCount = trPtr->numberOfPoints;
			DFUSsndPntPtr spPtr = (DFUSsndPntPtr) ((long)trPtr + DFUStrajectorySize);
			while (pointCount--) {
				spPtr->x = ntohl (spPtr->x);
				spPtr->y = ntohl (spPtr->y);
				spPtr->diameter = ntohl (spPtr->diameter);
				spPtr->transitionTime = ntohl (spPtr->transitionTime);
				spPtr->pauseTime = ntohl (spPtr->pauseTime);
				spPtr++;
				}
			trPtr = (DFUStrajPtr) spPtr;
			}
		dPtr = (DFUSdiffPtr) trPtr;
		}
}
/*************************************************************************************/
//set byte order for DIFF chunk when opening show
void fileToHostDIFF (DIFFchunkHdrPtr dcPtr) {
	dcPtr->numberOfSpkrMaps = ntohs (dcPtr->numberOfSpkrMaps);
	int spkrMapCount = dcPtr->numberOfSpkrMaps;
	dcPtr->numberOfTrajectories = ntohs (dcPtr->numberOfTrajectories);
	int trajectoryCount = dcPtr->numberOfTrajectories;
	DFUSspkrMapPtr smPtr = (DFUSspkrMapPtr) ((long)dcPtr + DIFFchunkHdrSize);
	while (spkrMapCount--) {
		smPtr->originX = ntohl (smPtr->originX);
		smPtr->originY = ntohl (smPtr->originY);
		smPtr->numberOfSpkrs = ntohl (smPtr->numberOfSpkrs);
		int spkrCount = smPtr->numberOfSpkrs;
		DFUSspkrPtr sPtr = (DFUSspkrPtr) ((long)smPtr + DFUSspeakerMapSize);
		while (spkrCount--) {
			sPtr->outputCh = ntohl (sPtr->outputCh);
			sPtr->x = ntohl (sPtr->x);
			sPtr->y = ntohl (sPtr->y);
			sPtr++;
			}
		smPtr = (DFUSspkrMapPtr) sPtr;
		}
	DIFFtrajPtr trPtr = (DIFFtrajPtr) smPtr;
	while (trajectoryCount--) {
		trPtr->cueNumber = ntohl (trPtr->cueNumber);
		trPtr->cueSubNumber = ntohl (trPtr->cueSubNumber);
		trPtr->inputCh = ntohs (trPtr->inputCh);
		trPtr->numberOfPoints = ntohs (trPtr->numberOfPoints);
		trPtr->listNumber = ntohs (trPtr->listNumber);
		trPtr->pathNumber = ntohs (trPtr->pathNumber);
		int pointCount = trPtr->numberOfPoints;
		DFUSsndPntPtr spPtr = (DFUSsndPntPtr) ((long)trPtr + DIFFtrajectorySize);
		while (pointCount--) {
			spPtr->x = ntohl (spPtr->x);
			spPtr->y = ntohl (spPtr->y);
			spPtr->diameter = ntohl (spPtr->diameter);
			spPtr->transitionTime = ntohl (spPtr->transitionTime);
			spPtr->pauseTime = ntohl (spPtr->pauseTime);
			spPtr++;
			}
		trPtr = (DIFFtrajPtr) spPtr;
		}
}
/*************************************************************************************/
//sets file name, path, number, type, modification date and date written to AB disk.
// srcPathName is the full path name of the source file, e.g. /audio/pianoJazz.aiff
// fileName is the name of the possibly converted file, e.g. pianoJazz-L.aiff. It can
// optionally be prefixed with path data, which will be stripped off before putting into
// the file data structure (ab64FileData).
// filenumber is 1..hvseln for audio, 1..hvshn for show
// type is ftAudio or ftShow
// modDate and transferTime are specified in seconds since 0 hours, 0 minutes, 0 seconds,
// January 1, 1970, Coordinated Universal Time, see POSIX 'time' function
void setFileData (char *fileName, char *srcPathName, long fileNumber,
	long fileType, long modTime, long transferTime) {

	struct {
		unsigned char code;
		char unused;
		short length;
		long fnumber;
		long ftype;
		struct ab64FileData fd;
		} setFileDataEN;

	memset (&setFileDataEN, 0 , sizeof (setFileDataEN)); // zero structure
	setFileDataEN.code = 0x81;
	setFileDataEN.length = htons (268);
	setFileDataEN.fnumber = htonl (fileNumber);
	setFileDataEN.ftype = htonl (fileType);
	//convert srcPathName to pure ASCII for transmission to AB64
	int fpnlen = strlen (srcPathName);
	char *cptr = srcPathName;
	while (fpnlen--) {
		if (isprint (*cptr) == 0)
			*cptr = '?';
		cptr++;
		}
	//copy path name into sourcePathName field, truncate from left if necessary
	int startCh = strlen (srcPathName) - (srcPathNameStringLen - 1);
	if (startCh < 0) startCh = 0;
	strcpy (setFileDataEN.fd.sourcePathName, srcPathName + startCh);

	//set file name, without path and without extension
	char *firstChar = strrchr (fileName, '/');
	if (firstChar == NULL)
		firstChar = fileName;
	else
		firstChar++;
	char *exten = strrchr (fileName, '.');
	if (exten == NULL)
		exten = strchr (fileName, 0);
	int len = exten - firstChar;
	if (len > 31) len = 31;
	memcpy (setFileDataEN.fd.filename, firstChar, len);
	setFileDataEN.fd.dlDate = htonl (transferTime - ABBDay);
	setFileDataEN.fd.modDate = htonl (modTime - ABBDay);
	sendto (dataSock, (const char*)&setFileDataEN, 268, 0, ENadrs, ENadrsSize);
}
/*************************************************************************************/
// Audio and show file transfer to AB64.
// Audio file must be AIFF, monophonic, 16 bit, 48 kHz.
// filePath is the full path name of the file to transfer, e.g. "/audio/Jazzpiano-L.aiff"
// sourcePath is the path name of the file that the user selected to start the transfer.
// This file may have been converted and/or split into mono files, if not,
// filePath and sourcePath should be the same string. If the file was converted and/or
// split, path for the original file is in sourcePath, and the path for the converted file is
// in filePath.
// File data written to AB64: sourcePathName: sourcePath
//							filename: filePath with path stripped off
// fileNumber is 1..hvseln for audio, 1..hvshn for show
// fileType is ftAudio or ftShow
int transferFileToAB64 (char *filePath, char *sourcePath, long fileNumber,
	long fileType) {

	if (fileTransferActive) {
		logError ("transferFileToAB64 called when another transfer was in progress");
		return false;
		}
	fileTransferActive = true;

	//make local copy of filePath and sourcePath, truncate from left if necessary
	int startCh = strlen (filePath) - (srcPathNameStringLen - 1);
	if (startCh < 0) startCh = 0;
	strcpy (fName, filePath + startCh);
	startCh = strlen (sourcePath) - (srcPathNameStringLen - 1);
	if (startCh < 0) startCh = 0;
	strcpy (sPath, sourcePath + startCh);
	//make local copy of fileNumber and fileType
	fNumber = fileNumber;
	fType = fileType;

	// open file
	fp = fopen (filePath, "rb");
	if (fp == NULL) {
		sprintf (astr, "error opening file %s", filePath);
		logError (astr);
		return false;
		}

	//get file size & modification time
	struct stat fileinfo;
	int x = stat (filePath, &fileinfo);
	totalTransferSize = fileinfo.st_size;
	mTime = fileinfo.st_mtime;	//calendar time when file was last modified
	struct tm *tmptr = localtime (&mTime);
#ifdef _FKN_WINDOWS
	 _tzset();  //set system time globals -- _timezone is seconds offset from UTC
	mTime -= _timezone;	
#else
	mTime += tmptr->tm_gmtoff;	//adjust to local time
#endif



	if (x != 0) {
		sprintf (astr, "error getting file info %s", filePath);
		logError (astr);
		return false;
		}

	bytesTransferred = 0;
	bytesReported = 0;

	if (fileType == ftAudio) {
		struct {
			long chunkID;
			long fSize;
			long formType;
			} fileHdr;

		//verify file header
		if (fread (&fileHdr, sizeof (fileHdr), 1, fp) != 1) {
			logError ("error reading audio file header");
			return false;
			}
		if (ntohl (fileHdr.chunkID) != 0x464F524D)	{	//'FORM'
			logError ("Unsupported file format");
			return false;
			}
		if ((ntohl (fileHdr.formType) != 0x41494646)	//'AIFF'
			&& (ntohl (fileHdr.formType) != 0x41494643)) {	//'AIFC'
			logError ("Unsupported file format");
			return false;
			}
		// scan file for 'COMM' chunk
		int fileOffset = sizeof (fileHdr);
		struct chunkHdr {
			long chunkID;
			long chunkSize;
			};
		struct {
			long chunkID;
			long chunkSize;
			short audioChannels;
			short framesHi;
			short framesLo;
			short bitsPerSample;
			long sampleRate;
			} commChunk;
		do {
			if (fread (&commChunk, sizeof (commChunk), 1, fp) != 1) {
				logError ("error reading file chunk");
				return false;
				}
			//set read position to start of next chunk
			fileOffset += ntohl (commChunk.chunkSize) + sizeof (struct chunkHdr);
			fseek (fp, fileOffset, SEEK_SET);
			} while (ntohl (commChunk.chunkID) != 0x434F4D4D); //'COMM'
		if (ntohs (commChunk.audioChannels) != 1) {
			logError ("audio file must be single channel");
			return false;
			}
		if (ntohl (commChunk.sampleRate) != 0x400EBB80) {	//48kHz in 'extended' float format
			logError ("audio file must be 48 kHz sample rate");
			return false;
			}
		if (ntohs (commChunk.bitsPerSample) != 16) {
			logError ("audio file must be 16 samples");
			return false;
			}
		// scan file for 'SSND' chunk - audio samples
		struct {
			long chunkID;
			long chunkSize;
			long offset;
			long blocksize;
			} ssndChunk;
		while (1) {
			if (fread (&ssndChunk, sizeof (ssndChunk), 1, fp) != 1) {
				logError ("error reading file chunk");
				return false;
				}
			if (ntohl (ssndChunk.chunkID) == 0x53534E44) //'SSND'
				break;
			//set read position to start of next chunk
			fileOffset += ntohl (ssndChunk.chunkSize) + sizeof (struct chunkHdr);
			fseek (fp, fileOffset, SEEK_SET);
			}	//end while
		totalTransferSize  = ntohl (ssndChunk.chunkSize);
		}	//end audio file
	else {	//show file
		struct abshowhdrv4 showHdr;
		if (fread (&showHdr, sizeof (showHdr), 1, fp) != 1) {
			logError ("error reading show file header");
			return false;
			}
		if (ntohl (showHdr.fileID) != 0x41425348) { //'ABSH'
			logError ("not a valid show file");
			return false;
			}
		fseek (fp, 0, SEEK_SET);
		}

	//get number of 64K segments, rounded up, send 'setup store'
	totalSegments = (int)((totalTransferSize + 65535) / 65536);
	autoUpdateOff ();
	sendSetupStore (fileNumber, fileType, totalSegments);

	//allocate file packets
	tfBuf = (unsigned char*)malloc(64 * 1032);		
	
	return true;
}
//*************************************************************************************
//send Setup Store command to AB64
void sendSetupStore (long fileNumber, long fileType, long segments) {
	long setupStoreEN[4];
	setupStoreEN[0] = htonl (0x6D000010);
	setupStoreEN[1] = htonl (fileNumber);
	setupStoreEN[2] = htonl (fileType);
	setupStoreEN[3] = htonl (segments);
	sendto (dataSock, (const char*)&setupStoreEN[0], 16, 0, ENadrs, ENadrsSize);		
//	sprintf (astr, "Setup Store sent, file %d, segments %d", fileNumber, segments);
//	logEvent (astr);
}
//*************************************************************************************
void cancelTransfer (void) {
	sendSetupStore (0, 0, 0);
	autoUpdateOn ();
	bytesTransferred = 0;
	bytesReported = 0;
	fileTransferActive = false;
};
//*************************************************************************************
void setupStoreReceived (int response) {
time_t transferTime;

	switch (response) {
	case 1:
	case 10:	//OK response
		//set file data in AB64

		time (&transferTime);	//get present coordinated universal time, seconds
#ifndef _FKN_WINDOWS		
		struct tm *tmptr = localtime (&transferTime);
		transferTime += tmptr->tm_gmtoff;	//adjust to local time
#endif	
		
		setFileData (fName, sPath, fNumber, fType, mTime, transferTime);
		//initialize and begin transfer
		transferringSegment = 0;
		fileTransferTimer = transferAckTimeout;
		packetNumber = 0;
		nextSegmentToAB64 (1);
		return;
	case 3:
		sprintf (astr, "file transfer already in progress");
		logError (astr);
		cancelTransfer();
		return;
	case 4:
	case 5:
		sprintf (astr, "not enough room on disk");
		logError (astr);
		cancelTransfer();
		return;
	case 7:
		return;	//store cancelled
	default:
		sprintf (astr, "bad setup store response %d", response);
		logError (astr);
		autoUpdateOn ();
		return;
	}
}
//*************************************************************************************
//transfer next segment (64 packet burst) to AB64
void nextSegmentToAB64 (long ack) {

//		sprintf (astr, "next segment %d, ack %d", packetNumber, ack);
//		logEvent (astr);

	if (transferringSegment >= totalSegments && ack == 1) {
		bytesTransferred = 0;	//all segments transferred
		autoUpdateOn ();
		free (tfBuf);
		fileTransferActive = false;
		return;
		}

	int x;
	int bytesRead;
	int padBytes;
	unsigned char *cptr;
	if (ack == 1)	{	//last burst was ACKed, refill all packets
		fileTransferTimer = transferAckTimeout;
		transferringSegment++;
		cptr = tfBuf;
		x = 64;
		while (x--) {
			*((long *)cptr) = htonl (0x84000408);
			cptr += 4;
			*((long *)cptr) = htonl (packetNumber++);
			cptr += 4;
			bytesRead = fread (cptr, 1, 1024, fp);
			bytesTransferred += bytesRead;
			cptr += bytesRead;
			padBytes = 1024 - bytesRead;
			while (padBytes--)
				*cptr++ = 0;
			}
		}

	//transmit packets
	cptr = tfBuf;
	x = 64;
	while (x--) {
//		if (testDrop == 0 && x == 44) {
//			testDrop = 1;
//			cptr += 1032;
//			continue;
//			}
		sendto (dataSock, (const char*)cptr, 1032, 0, ENadrs, ENadrsSize);   
//		sprintf (astr, "%d sent", x);
//		logEvent (astr);
		cptr += 1032;
		}
}
//*************************************************************************************
//send erase file command to AB64
void eraseAB64File (long fileNumber, long fileType) {
	long eraseFileEN[3];
	eraseFileEN[0] = htonl (0x6A00000C);
	eraseFileEN[1] = htonl (fileNumber);
	eraseFileEN[2] = htonl (fileType);
	sendto (dataSock, (const char*)&eraseFileEN[0], 12, 0, ENadrs, ENadrsSize);   
//	sprintf (astr, "file erased, file %d, type %d", fileNumber, fileType);
//	logEvent (astr);
}
/*************************************************************************************/
// Audio and show file transfer from AB64.
// filePath is the full path name for the file, without extension, e.g. "/audio/preshow/Jazzpiano"
// audio files will have ".aiff" extension added
// show files will have ".sho" extension added
// file will be created if it does not exist, overwritten if it does
// fileNumber is 1..hvseln for audio, 1..hvshn for show
// fileType is ftAudio or ftShow
// if fileType is ftAudio, channels is number of channels in file 1..64, not used for show file
// blocks is file size in blocks
int transferFileFromAB64 (char *filePath, long fileNumber, long fileType, long channels, long blocks) {

	if (fileTransferActive) {
		logError ("transferFileFromAB64 called when another transfer was in progress");
		return false;
		}
	fileTransferActive = true;

	//send request to AB64
	autoUpdateOff ();
	transferringSegment = 0;
	uploadAudioFileEN[0] = ntohl (0x87000010);
	uploadAudioFileEN[1] = ntohl (fileNumber);
	uploadAudioFileEN[2] = ntohl (fileType);
	uploadAudioFileEN[3] = ntohl (transferringSegment);
	sendto (dataSock, (const char*)&uploadAudioFileEN[0], 16, 0, ENadrs, ENadrsSize);   

	totalSegments = ((blocks * 512) / 32768) * 32;
	totalTransferSize = totalSegments * 1024;
	bytesTransferred = 0;
	bytesReported = 0;

	// open file for write
	char *fpstr = (char*)malloc (strlen (filePath) + 6);   
	strcpy (fpstr, filePath);
	if (fileType == ftAudio)
		strcat (fpstr, ".aiff");
	else
		strcat (fpstr, ".sho");
	fp = fopen (fpstr, "wb");
	free (fpstr);
	if (fp == NULL) {
		sprintf (astr, "error opening file %s", filePath);
		logError (astr);
		return false;
		}

	channelCount = channels;
	if (fileType == ftAudio) {
		if (channels == 0) {
			logError ("transfer from AB64: audio channels == 0");
			return false;
			}
		struct aiffHeader {
			long formID;
			long totalCkSize;
			long formType;
			long commCkID;
			long commDataSize;
			short numChannels;
			unsigned short numSampleFramesHi;	//for proper structure packing
			unsigned short numSampleFramesLo;	//(a long here could result in a padding word)
			short sampleSize;
			unsigned char sampleRate[12];			//80 bit IEEE float, 10 bytes, plus two padding bytes
			long sndCkID;
			long sndDataSize;
			long offset;
			long blocksize;
			} aiffHdr;

		memset (&aiffHdr, 0, sizeof (aiffHdr));	//zero the structure
		int audioBytes = blocks * 512;
		aiffHdr.formID = htonl (0x464F524D);			//'FORM'
		aiffHdr.totalCkSize = htonl (audioBytes + sizeof (aiffHdr) - 8);
		aiffHdr.formType = htonl (0x41494646);		//'AIFF'
		aiffHdr.commCkID = htonl (0x434F4D4D);	//'COMM'
		aiffHdr.commDataSize = htonl (20);
		aiffHdr.numChannels = htons ((short)channels);		
		int sampleFrames = (audioBytes / 2) / channels;
		aiffHdr.numSampleFramesHi = htons (sampleFrames >> 16);
		aiffHdr.numSampleFramesLo = htons (sampleFrames & 0xFFFF);
		aiffHdr.sampleSize = htons (16);
		aiffHdr.sampleRate[0] = 0x40;		//48000 in 80 bit IEEE float, remaining bytes are zero
		aiffHdr.sampleRate[1] = 0x0E;
		aiffHdr.sampleRate[2] = 0xBB;
		aiffHdr.sampleRate[3] = 0x80;
		aiffHdr.sndCkID = htonl (0x53534E44);		//'SSND'
		aiffHdr.sndDataSize = htonl (audioBytes + 8);
		int bytesWritten = fwrite (&aiffHdr, 1, sizeof (aiffHdr), fp);
		if (bytesWritten != sizeof (aiffHdr)) {
			sprintf (astr, "error writing AIFF file header %s", filePath);
			logError (astr);
			return false;
			}
		}	//end audio file

	//allocate transfer buffer, 8K per channel if audio, 32K buffer if show
	if (fileType == ftAudio)
		tfBuf = (unsigned char*)malloc (channels * 8 * 1024);
	else
		tfBuf = (unsigned char*)malloc (32 * 1024);
	tfBufPtr = tfBuf;
	packetsInBuffer = 0;
	return true;
}
//*************************************************************************************
// handles incoming file packets
// audio file interleave: 8K (8 packets) per channel
void nextSegmentFromAB64 (getFilePtr gfPtr) {

	#if defined(__LITTLE_ENDIAN__)
	gfPtr->fileNumber = ntohl (gfPtr->fileNumber);
	gfPtr->fileType = ntohl (gfPtr->fileType);
	gfPtr->segment = ntohl (gfPtr->segment);
	gfPtr->returncode = ntohl (gfPtr->returncode);
	#endif

	if (gfPtr->returncode != 1) {
		sprintf (astr, "GET FILE response %d", gfPtr->returncode);
		logError (astr);
		return;
		}

	if (gfPtr->segment != transferringSegment) {	//not the expected segment
		//request the previous burst be sent again
		sendto (dataSock, (const char*)&uploadAudioFileEN[0], 16, 0, ENadrs, ENadrsSize);		
		return;
		}

	//copy incoming bytes to buffer
	transferringSegment++;
	memcpy (tfBufPtr, &gfPtr->fileBytes, 1024);
	tfBufPtr += 1024;
	bytesTransferred += 1024;
	packetsInBuffer++;
	int objectsWritten;
	if (gfPtr->fileType == ftAudio && packetsInBuffer == 8 * channelCount) { //time to write buffer to file
		if (channelCount == 1) {	//mono file, no need to interleave
			objectsWritten = fwrite (tfBuf, 1, 8*1024, fp);
			if (objectsWritten != 8*1024) {
				sprintf (astr, "error writing audio file");
				logError (astr);
				return;
				}
			} //end mono file
		else {	//mulitchannel file
			unsigned char *mPtr = tfBuf;
			int x = 8*1024/2;	//sample frames for each channel
			while (x--) {
				int ch = channelCount;
				unsigned char *sPtr = mPtr;
				while (ch--) {
					objectsWritten = fwrite (sPtr, 2, 1, fp);
					if (objectsWritten != 1) {
						sprintf (astr, "error writing audio file");
						logError (astr);
						return;
						}
					sPtr += 8*1024;	//next 8K channel buffer
					}
				mPtr += 2;
				}
			}	//end multichannel file
		tfBufPtr = tfBuf;
		packetsInBuffer = 0;
		}	//end time to write buffer to audio file

	if (gfPtr->fileType == ftShow && packetsInBuffer == 32) { //time to write buffer to file
			objectsWritten = fwrite (tfBuf, 1, 32*1024, fp);
			if (objectsWritten != 32*1024) {
				sprintf (astr, "error writing show file");
				logError (astr);
				return;
				}
			tfBufPtr = tfBuf;
			packetsInBuffer = 0;
			}

	if (transferringSegment >= totalSegments) {
		bytesTransferred = 0;	//all segments transferred
		fromCmplFlag = true;
		fileTransferActive = false;
		autoUpdateOn ();
		free (tfBuf);
		fclose (fp);
		return;
		}

	if (transferringSegment < totalSegments && transferringSegment % 32 == 0) { //request next burst of 32 packets
		uploadAudioFileEN[3] = ntohl (transferringSegment);
		sendto (dataSock, (const char*)&uploadAudioFileEN[0], 16, 0, ENadrs, ENadrsSize);		
		}
}
//*************************************************************************************
int transferFirmware (char *filePath, int writeToDisk) {

	if (fileTransferActive) {
		logError ("transferFirmware called when another transfer was in progress");
		return false;
		}

	// open file
	fp = fopen (filePath, "rb");
	if (fp == NULL) {
		sprintf (astr, "error opening file %s", filePath);
		logError (astr);
		return false;
		}

	fileTransferActive = true;

	pktNumber = 0;
	lastAck = false;
	tfBuf = (unsigned char*)malloc (1050);

	//clear out the socket
	autoUpdateOff ();
	millisleep (50);	//wait for auto updates to end
	fd_set readfds;
	int dataReady;
	while (1) {
		FD_ZERO (&readfds);
		FD_SET ((unsigned int)dataSock, &readfds);   
		dataReady = select (dataSock+1, &readfds, NULL, NULL, &timeout);
		if (dataReady)
			recvfrom (dataSock, (char*)recvPacketBuf, packetBufSize, 0, NULL, 0);
		else
			break;
		}

	struct {
		unsigned short code;
		short length;
		long packetNumber;
		long responseCode;
		} reply;

	while (1) {
		long *lptr = (long *)tfBuf;
		*lptr++ = htonl (0x6B00040C);
		*lptr++ = htonl (pktNumber);
		*lptr++ = 0;	//data packet
		int bytesRead = fread (lptr, 1, 1024, fp);
		if (bytesRead < 1024) {	//last packet, pad, set run or write command
			lastAck = true;
			unsigned char *cptr = (unsigned char *)lptr + bytesRead;
			int padBytes = 1024 - bytesRead;
			while (padBytes--)
				*cptr++ = 0;
			lptr = (long *)(tfBuf + 8);
			if (writeToDisk)
				*lptr = htonl (2);	//write new firmware
			else
				*lptr = htonl (1);	//run new firmware
			sendto (dataSock, (const char*)tfBuf, 1036, 0, ENadrs, ENadrsSize);
			autoUpdateOn ();
			fclose (fp);   //changed from open for windoz
			free (tfBuf);
			fileTransferActive = false;
			fwCmplFlag = true;
			return true;
			}	//end last packet


		sendto (dataSock, (const char*)tfBuf, 1036, 0, ENadrs, ENadrsSize);

		int retry = 20;
		while (retry--) {
			FD_ZERO (&readfds);
			FD_SET ((unsigned int)dataSock, &readfds);
			dataReady = select (dataSock+1, &readfds, NULL, NULL, &timeout);
			if (dataReady) {
				recvfrom (dataSock, (char*)&reply, sizeof (reply), 0, NULL, 0);   
				reply.code  = ntohs (reply.code);
				reply.length  = ntohs (reply.length);
				reply.packetNumber  = ntohl (reply.packetNumber);
				reply.responseCode  = ntohl (reply.responseCode);
				if (reply.responseCode == 0) {
					logError ("firmware download NAK received");
					return false;
					}
				else if (reply.packetNumber != pktNumber) {
					sprintf (astr, "firmware download packet number mismatch %d %d",
						reply.packetNumber, pktNumber);
					logError (astr);
					return false;
					}
				else
					break;		//exit retry loop
				}	//end data ready
			else
				millisleep (3);
			}	//end retry loop

		pktNumber++;

	}	//end while
}
/*************************************************************************************/
/*** AB64 Command sending functions
/*************************************************************************************/
void setAB64Name (char *name) {
	struct {
		short code;
		short length;
		char nameString[32];
		} ABName;
	ABName.code = htons (0x8600);
	ABName.length = htons (36);
	int nameLen = strlen (name);
	if (nameLen > 31) nameLen = 31;
	memcpy (&ABName.nameString, name, nameLen);
	ABName.nameString[nameLen] = 0;
	sendto (dataSock, (const char*)&ABName, 36, 0, ENadrs, ENadrsSize);   
}
/*************************************************************************************/
void getDiskInfo (void) {
	sendto (dataSock, (const char*)&requestDiskInfo, 4, 0, ENadrs, ENadrsSize);   
}
/*************************************************************************************/
//  translates a channel number (0..63 for inputs, 64..127 for outputs)
//	into a command channel number. Also translates back the other way.
//  (Swaps bits 5 and 6.)
//	command channel number:
//      00H-1FH for inputs 1 to 32, 20H-3FH for outputs 1 to 32
//     40H-5FH for inputs 33 to 64, 60H-7FH for outputs 33 to 64
long translateChannel (long inch) {

long i;
	i = inch & 0x1F;	//clear bits 5, 6 & 7
	if (inch & 0x20)
		i |= 0x40;
	if (inch & 0x40)
		i |= 0x20;
	return i;
}
/*************************************************************************************/
void setInputLevel (long ch, long level, long exp, long hour, long minute, long second,
						long frame, long subframe, long framesPerSecond) {

	cptr = &MSC16[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0;
	*cptr++ = ch - 1;
	*cptr++ = (unsigned char)level;  
	switch (framesPerSecond) {
	case 25:
		*cptr++ = hour + 0x20;
		break;
	case 30:
		*cptr++ = hour + 0x60;
		break;
	case 24:
	default:
		*cptr++ = (unsigned char)hour;  
		break;
	}
	*cptr++ = minute + ((exp & 1) << 6);
	*cptr++ = (unsigned char)second;   
	*cptr++ = (unsigned char)frame;  
	*cptr++ = (unsigned char)subframe;  
	sendCommand (MSC16, 16);
}
/*************************************************************************************/
void setOutputLevel (long ch, long level, long exp, long hour, long minute, long second,
						long frame, long subframe, long framesPerSecond) {

	cptr = &MSC16[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 3;
	*cptr++ = ch - 1;
	*cptr++ = (unsigned char)level;  
	switch (framesPerSecond) {
	case 25:
		*cptr++ = hour + 0x20;
		break;
	case 30:
		*cptr++ = hour + 0x60;
		break;
	case 24:
	default:
		*cptr++ = (unsigned char)hour;  
		break;
	}
	*cptr++ = minute + ((exp & 1) << 6);
	*cptr++ = (unsigned char)second;  
	*cptr++ = (unsigned char)frame;  
	*cptr++ = (unsigned char)subframe;  
	sendCommand (MSC16, 16);
}
/*************************************************************************************/
void setCrosspointLevel (long inch, long ouch, long level, long exp,
						 long hour, long minute, long second, long frame, long subframe, long framesPerSecond) {

	cptr = &MSC16[5];
	*cptr++ = 6;
	*cptr++ = 1;
	*cptr++ = inch - 1;
	*cptr++ = ouch - 1;
	*cptr++ = (unsigned char)level;  
	switch (framesPerSecond) {
	case 25:
		*cptr++ = hour + 0x20;
		break;
	case 30:
		*cptr++ = hour + 0x60;
		break;
	case 24:
	default:
		*cptr++ = (unsigned char)hour;  
		break;
	}
	*cptr++ = minute + ((exp & 1) << 6);
	*cptr++ = (unsigned char)second;  
	*cptr++ = (unsigned char)frame;  
	*cptr++ = (unsigned char)subframe;  
	sendCommand (MSC16, 16);
}
/*************************************************************************************/
void setCrosspointPhase (long inch, long ouch, long phase) { //phase 0: no inversion, 1: 180 degree inversion

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0x08;
	*cptr++ = inch - 1;
	*cptr++ = ouch - 1;
	*cptr++ = (unsigned char)phase;  
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
void setCrosspointDelay (long inch, long ouch, long dly,
		 long hour, long minute, long second, long frame, long subframe, long framesPerSecond) {

	cptr = &sysex19[7];

	*cptr++ = 0x16;
	*cptr++ = inch - 1;
	*cptr++ = ouch - 1;
	*cptr++ = dly & 0x7F;
	*cptr++ = (dly >> 7) & 0x7F;
	*cptr++ = (dly >> 14) & 0x7F;

	switch (framesPerSecond) {
	case 25:
		*cptr++ = hour + 0x20;
		break;
	case 30:
		*cptr++ = hour + 0x60;
		break;
	case 24:
	default:
		*cptr++ = (unsigned char)hour;  
		break;
	}
	*cptr++ = (unsigned char)minute;  
	*cptr++ = (unsigned char)second;  
	*cptr++ = (unsigned char)frame;  
	*cptr++ = (unsigned char)subframe;  
	sendCommand (sysex19, 19);
}
/*************************************************************************************/
void setSourceMix (long ch, long analog, long digital, long playback) {

	cptr = &sysex13[7];
	*cptr++ = 0x1A;
	*cptr++ = ch - 1;
	*cptr++ = (unsigned char)analog;  
	*cptr++ = (unsigned char)playback;  
	*cptr++ = (unsigned char)digital;  
	sendCommand (sysex13, 13);
}
/*************************************************************************************/
void setInputDelay (long ch, long dly) {	// dly in samples

	cptr = &sysex13[7];
	*cptr++ = 0x1B;
	*cptr++ = (unsigned char)translateChannel (ch-1);  
	*cptr++ = dly & 0x7F;
	*cptr++ = (dly >> 7) & 0x7F;
	*cptr++ = (dly >> 14) & 0x7F;
	sendCommand (sysex13, 13);
}
/*************************************************************************************/
void setOutputDelay (long ch, long dly)	 { // dly in samples

	cptr = &sysex13[7];
	*cptr++ = 0x1B;
	*cptr++ = (unsigned char)translateChannel (ch+63);  
	*cptr++ = dly & 0x7F;
	*cptr++ = (dly >> 7) & 0x7F;
	*cptr++ = (dly >> 14) & 0x7F;
	sendCommand (sysex13, 13);
}
/*************************************************************************************/
void clearAllDelays (void) {

	cptr = &sysex12[7];
	*cptr++ = 0x0C;
	*cptr++ = 0x43;
	*cptr++ = 0x1B;
	*cptr++ = 0x68;
	sendCommand (sysex12, 12);
}
/*************************************************************************************/
void setInputDelayEnable (long ch, long enable)	{ // enable: 1, disable: 0

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0x0B;
	*cptr++ = (unsigned char)translateChannel (ch-1);  
	*cptr++ = enable & 1;
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
void setOutputDelayEnable (long ch, long enable) {	// enable: 1, disable: 0

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0x0B;
	*cptr++ = (unsigned char)translateChannel (ch+63);  
	*cptr++ = enable & 1;
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
void setInputMute (long ch, long mute)	{ // mute: 1, unmute: 0

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0x07;
	*cptr++ = ch-1;
	*cptr++ = mute & 1;
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
void setOutputMute (long ch, long mute) {	// mute: 1, unmute: 0

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0x08;
	*cptr++ = ch-1;
	*cptr++ = mute & 1;
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
void setInputSolo (long ch, long solo) {	// solo: 1, unsolo: 0

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0x09;
	*cptr++ = ch-1;
	*cptr++ = solo & 1;
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
void setOutputSolo (long ch, long solo) { // solo: 1, unsolo: 0

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0x0A;
	*cptr++ = ch-1;
	*cptr++ = solo & 1;
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
void pushFrontPanelButton (long button) {

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0x1E;
	*cptr++ = (unsigned char)button;  
	*cptr++ = 0;
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
void disablePlaybackChannel (long plbkch) {

	cptr = &playbackStr[5];
	*cptr++ = 0x1C;
	cptr += sprintf ((char *)cptr, "%ld", plbkch-1);
	*cptr++ = 0xf7;
	sendCommand (playbackStr, cptr - playbackStr);
}
/*************************************************************************************/
void locationString (long sel, long min, long sec, long frame, long framesPerSecond) {

	cptr += sprintf ((char *)cptr, "%ld", sel);
	if (min | sec | frame) {
		*cptr++ = '.';
		cptr += sprintf ((char *)cptr, "%ld", min);
		if (sec | frame) {
			*cptr++ = '.';
			cptr += sprintf ((char *)cptr, "%ld", sec);
			if (frame) {
				*cptr++ = '.';
				cptr += sprintf ((char *)cptr, "%ld", frame * 48000 / framesPerSecond);
			}
		}
	}
}
/*************************************************************************************/
void setPlaybackLocation (long plbkch, long sel, long min, long sec, long frame, long framesPerSecond) {

	cptr = &playbackStr[5];
	*cptr++ = 0x05;
	locationString (sel, min, sec, frame, framesPerSecond);
	*cptr++ = 0;
	cptr += sprintf ((char *)cptr, "%ld", plbkch-1);
	*cptr++ = 0xf7;
	sendCommand (playbackStr, cptr - playbackStr);
}
/*************************************************************************************/
void setPlaybackBlockLocation (long plbkch, long sel, long block) {

	setPBLKLocation (5, plbkch, sel, block);
}
/*************************************************************************************/
void setStopBlockLocation (long plbkch, long sel, long block) {

	setPBLKLocation (2, plbkch, sel, block);
}
/*************************************************************************************/
void setResumeBlockLocation (long plbkch, long sel, long block) {

	setPBLKLocation (3, plbkch, sel, block);
}
/*************************************************************************************/
void setPBLKLocation (long cmdCode, long plbkch, long sel, long block) {

long samples, seconds, minutes;

	minutes = (long)(block / blocksPerMinute);  
	block -= (long)(minutes * blocksPerMinute);
	seconds = (long)(block / blocksPerSecond);
	block -= (long)(seconds * blocksPerSecond);
	samples = block * 256;

	cptr = &playbackStr[5];
	*cptr++ = (unsigned char)cmdCode;   
	cptr += sprintf ((char *)cptr, "%ld", sel);
	if (minutes | seconds | samples) {
		*cptr++ = '.';
		cptr += sprintf ((char *)cptr, "%ld", minutes);
		if (seconds | samples) {
			*cptr++ = '.';
			cptr += sprintf ((char *)cptr, "%ld", seconds);
			if (samples) {
				*cptr++ = '.';
				cptr += sprintf ((char *)cptr, "%ld", samples);
				}
			}
		}
	*cptr++ = 0;
	cptr += sprintf ((char *)cptr, "%ld", plbkch-1);
	*cptr++ = 0xf7;
	sendCommand (playbackStr, cptr - playbackStr);
}
/*************************************************************************************/
void goPlayback (long plbkch, long sel, long min, long sec, long frame, long framesPerSecond) {

	cptr = &playbackStr[5];
	*cptr++ = 0x01;
	locationString (sel, min, sec, frame, framesPerSecond);
	*cptr++ = 0;
	cptr += sprintf ((char *)cptr, "%ld", plbkch-1);
	*cptr++ = 0xf7;
	sendCommand (playbackStr, cptr - playbackStr);
}
/*************************************************************************************/
//bits set in chSet1 & chSet2 determine which channels start playback
//ch1 = chSet1:bit0, ch32 = chSet1:bit31, ch33 = chSet2:bit0, ch64 = chSet2:bit31
void goPlaybackChannels (long chSet1, long chSet2)  {

	cptr = &sysex19[7];
	*cptr++ = 0x03;
	*cptr++ = chSet1 & 0x7F;			//0..6
	*cptr++ = (chSet1 >> 7) & 0x7F;	//7..13
	*cptr++ = (chSet1 >> 14) & 0x7F;	//14..20
	*cptr++ = (chSet1 >> 21) & 0x7F;	//21..27
	*cptr++ = ((chSet1 >> 28) & 0xF) | ((chSet2 & 0x7) << 4);	//28..34
	*cptr++ = (chSet2 >> 3) & 0x7F;	//35..41
	*cptr++ = (chSet2 >> 10) & 0x7F;	//42..48
	*cptr++ = (chSet2 >> 17) & 0x7F;	//49..55
	*cptr++ = (chSet2 >> 24) & 0x7F;	//55..62
	*cptr++ = (chSet2 >> 31) & 1;		//63
	*cptr++ = 0xf7;
	sendCommand (sysex19, 19);
}
/*************************************************************************************/
void loadPlaybackMultiChannel (long plbkch, long sel, long min, long sec, long frame, long framesPerSecond) {

	cptr = &playbackMultiStr[5];
	*cptr++ = 0x05;
	locationString (sel, min, sec, frame, framesPerSecond);
	*cptr++ = 0;
	cptr += sprintf ((char *)cptr, "%ld", plbkch-1);
	*cptr++ = 0xf7;
	sendCommand (playbackMultiStr, cptr - playbackMultiStr);
}
/*************************************************************************************/
// inOut ==  0: record inputs, 1: record outputs
void setupRecord (long selection, long inOut)  {

	cptr = &MSC11[5];
	*cptr++ = 0x06;
	*cptr++ = 0x0A;
	*cptr++ = selection & 0x7F;			//selection number, low 7 bits
	*cptr++ = (selection >> 7) & 0x7F;	//selection number, high 7 bits
	*cptr++ = (unsigned char)inOut;  
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
// enable = 1 to enable channel, 0 to disable channel
void enableRecordChannel (long channel, long enable)  {

	cptr = &MSC11[5];
	*cptr++ = 0x06;
	*cptr++ = 0x00;
	*cptr++ = 0x21;
	*cptr++ = channel-1;
	*cptr++ = (unsigned char)enable;  
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
void goAllPlayback (void) {

	cptr = &playbackStr[5];
	*cptr++ = 0x01;
	*cptr++ = 0xf7;
	sendCommand (playbackStr, cptr - playbackStr);
}
/*************************************************************************************/
void playbackStop (long plbkch) {

	cptr = &playbackStr[5];
	*cptr++ = 0x02;
	*cptr++ = '0';
	*cptr++ = 0;
	cptr += sprintf ((char *)cptr, "%ld", plbkch-1);
	*cptr++ = 0xf7;
	sendCommand (playbackStr, cptr - playbackStr);
}
/*************************************************************************************/
void setPlaybackStop (long plbkch, long sel, long min, long sec, long frame, long framesPerSecond) {

	cptr = &playbackStr[5];
	*cptr++ = 0x02;
	locationString (sel, min, sec, frame, framesPerSecond);
	*cptr++ = 0;
	cptr += sprintf ((char *)cptr, "%ld", plbkch-1);
	*cptr++ = 0xf7;
	sendCommand (playbackStr, cptr - playbackStr);
}
/*************************************************************************************/
void stopAllPlayback (void) {

	cptr = &playbackStr[5];
	*cptr++ = 0x02;
	*cptr++ = 0xf7;
	sendCommand (playbackStr, cptr - playbackStr);
}
/*************************************************************************************/
void setPlaybackResume (long plbkch, long sel, long min, long sec, long frame, long framesPerSecond) {

	cptr = &playbackStr[5];
	*cptr++ = 0x03;
	locationString (sel, min, sec, frame, framesPerSecond);
	*cptr++ = 0;
	cptr += sprintf ((char *)cptr, "%ld", plbkch-1);
	*cptr++ = 0xf7;
	sendCommand (playbackStr, cptr - playbackStr);
}
/*************************************************************************************/
void turnOffMatrix (void) {

	sendCommand (matrixOffStr, 11);
}
/*************************************************************************************/
void allOnMatrix (void) {

	sendCommand (matrixAllOnStr, 11);
}
/*************************************************************************************/
void diagonalMatrix (void) {

	sendCommand (matrixDiagonalStr, 11);
}
/*************************************************************************************/
void dualDiagonalMatrix (void) {	//for testing AB1616 & 1616HD

	long i;
	turnOffMatrix ();
	// turn on all inputs
	MSC16[5] = 6;
	MSC16[6] = 0;
	MSC16[7] = 0;
	MSC16[9] = 0x7f;
	for (i = 0; i <=15; i++) {
		MSC16[8] = (unsigned char)i;  
		sendCommand (MSC16, 16);
	}
	/* turn on all outputs */
	MSC16[7] = 3;
	for (i = 0; i <=15; i++) {
		MSC16[8] = (unsigned char)i;  
		sendCommand (MSC16, 16);
	}
	/* turn on diagonal crosspoints (i , i) */
	MSC16[6] = 1;
	for (i = 0; i <=15; i++) {
		MSC16[7] = (unsigned char)i;  
		MSC16[8] = (unsigned char)i;  
		sendCommand (MSC16, 16);
	}
	/* also diagonal crosspoints (i , i+8) */
	for (i = 0; i <=7; i++) {
		MSC16[7] = (unsigned char)i;  
		MSC16[8] = i + 8;
		sendCommand (MSC16, 16);
	}
}
/*************************************************************************************/
void in1toAllMatrix (void) {

	long i;
	turnOffMatrix ();
	/* turn on input 0 */
	MSC16[5] = 6;
	MSC16[6] = 0;
	MSC16[7] = 0;
	MSC16[8] = 0;
	MSC16[9] = 0x7f;
	sendCommand (MSC16, 16);

	/* turn on all outputs */
	MSC16[6] = 0;
	MSC16[7] = 3;
	for (i = 0; i <=63; i++) {
		MSC16[8] = (unsigned char)i;  
		sendCommand (MSC16, 16);
	}

	/* turn on crosspoints (0 , i) */
	MSC16[6] = 0x01;
	for (i = 0; i <=63; i++) {
		MSC16[7] = 0;
		MSC16[8] = (unsigned char)i;  
		sendCommand (MSC16, 16);
	}
}
/*************************************************************************************/
void openList (long show, long list) {	//opens list, and show if show not already open

	cptr = &MSCStr[5];
	*cptr++ = 0x1B;
	cptr += sprintf ((char *)cptr, "%ld", show);
	*cptr++ = '.';
	cptr += sprintf ((char *)cptr, "%ld", list);
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void closeList (long list) {	//closes list

	cptr = &MSCStr[5];
	*cptr++ = 0x1C;
	cptr += sprintf ((char *)cptr, "%ld", list);
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void openPath (long list, long path)	{ //opens path, and list if list not already open

	cptr = &MSCStr[5];
	*cptr++ = 0x1D;
	cptr += sprintf ((char *)cptr, "%ld", list);
	*cptr++ = '.';
	cptr += sprintf ((char *)cptr, "%ld", path);
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void closePath (long list, long path)	{ //closes path

	cptr = &MSCStr[5];
	*cptr++ = 0x1E;
	cptr += sprintf ((char *)cptr, "%ld", list);
	*cptr++ = '.';
	cptr += sprintf ((char *)cptr, "%ld", path);
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void allOff (void) {

	cptr = &MSCStr[5];
	*cptr++ = 0x08;
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
 //sets ALL OFF action flags (see command set for bit definitions)
 void setAllOffActions (long flags) {

	cptr = &MSCStr[5];
	*cptr++ = 0x06;
	*cptr++ = 0;
	*cptr++ = 0x19;
	*cptr++ = (unsigned char)flags;  
	*cptr++ = 0;
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
 //sets RESET action flags (see command set for bit definitions)
void setResetActions (long flag1, long flag2) {

	cptr = &MSCStr[5];
	*cptr++ = 0x06;
	*cptr++ = 0;
	*cptr++ = 0x15;
	*cptr++ = (unsigned char)flag1;
	*cptr++ = (unsigned char)flag2;
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void restore (void) {

	cptr = &MSCStr[5];
	*cptr++ = 0x09;
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void reset (void) {

	cptr = &MSCStr[5];
	*cptr++ = 0x0A;
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void cueString (long cue, long sub1, long sub2, long sub3, long sub4) {

	cptr += sprintf ((char *)cptr, "%ld", cue);
	if (sub1 | sub2 | sub3 | sub4) {
		*cptr++ = '.';
		cptr += sprintf ((char *)cptr, "%ld", sub1);
		if (sub2 | sub3 | sub4) {
			*cptr++ = '.';
			cptr += sprintf ((char *)cptr, "%ld", sub2);
			if (sub3 | sub4) {
				*cptr++ = '.';
				cptr += sprintf ((char *)cptr, "%ld", sub3);
				if (sub4) {
					*cptr++ = '.';
					cptr += sprintf ((char *)cptr, "%ld", sub4);
				}
			}
		}
	}
}
/*************************************************************************************/
void goCue (long cue, long sub1, long sub2, long sub3, long sub4, long list, long path) {

	cptr = &MSCStr[5];
	*cptr++ = 0x01;
	cueString (cue, sub1, sub2, sub3, sub4);
	if (list) {
		*cptr++ = 0;
		cptr += sprintf ((char *)cptr, "%ld", list);
		if (path) {
		*cptr++ = 0;
		cptr += sprintf ((char *)cptr, "%ld", path);
		}
	}
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void goCueJamClock (long cue, long sub1, long sub2, long sub3, long sub4, long list, long path) {

	cptr = &MSCStr[5];
	*cptr++ = 0x10;
	cueString (cue, sub1, sub2, sub3, sub4);
	if (list) {
		*cptr++ = 0;
		cptr += sprintf ((char *)cptr, "%ld", list);
		if (path) {
		*cptr++ = 0;
		cptr += sprintf ((char *)cptr, "%ld", path);
		}
	}
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void loadCue (long cue, long sub1, long sub2, long sub3, long sub4, long list, long path) {

	cptr = &MSCStr[5];
	*cptr++ = 0x05;
	cueString (cue, sub1, sub2, sub3, sub4);
	if (list) {
		*cptr++ = 0;
		cptr += sprintf ((char *)cptr, "%ld", list);
		if (path) {
			*cptr++ = 0;
			cptr += sprintf ((char *)cptr, "%ld", path);
		}
	}
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void stopCue (long cue, long sub1, long sub2, long sub3, long sub4, long list) {

	cptr = &MSCStr[5];
	*cptr++ = 0x02;
	cueString (cue, sub1, sub2, sub3, sub4);
	if (list) {
		*cptr++ = 0;
		cptr += sprintf ((char *)cptr, "%ld", list);
		}
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void resumeCue (long cue, long sub1, long sub2, long sub3, long sub4, long list) {

	cptr = &MSCStr[5];
	*cptr++ = 0x03;
	cueString (cue, sub1, sub2, sub3, sub4);
	if (list) {
		*cptr++ = 0;
		cptr += sprintf ((char *)cptr, "%ld", list);
		}
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void incrementStandbyCue (long list) {

	cptr = &MSCStr[5];
	*cptr++ = 0x11;
	if (list)
		cptr += sprintf ((char *)cptr, "%ld", list);
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void decrementStandbyCue (long list) {

	cptr = &MSCStr[5];
	*cptr++ = 0x12;
	if (list)
		cptr += sprintf ((char *)cptr, "%ld", list);
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void startListClock (long list)	{ //start list clock

	cptr = &MSCStr[5];
	*cptr++ = 0x15;
	if (list)
		cptr += sprintf ((char *)cptr, "%ld", list);
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void stopListClock (long list)	{ //stop list clock

	cptr = &MSCStr[5];
	*cptr++ = 0x16;
	if (list)
		cptr += sprintf ((char *)cptr, "%ld", list);
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void setListClock (long hour, long minute, long second, long frame, long subframe, long list, long framesPerSecond) {

	cptr = &MSCStr[5];
	*cptr++ = 0x18;
	switch (framesPerSecond) {
	case 25:
		*cptr++ = hour + 0x20;
		break;
	case 30:
		*cptr++ = hour + 0x60;
		break;
	case 24:
	default:
		*cptr++ = (unsigned char)hour;
		break;
	}
	*cptr++ = (unsigned char)minute;  
	*cptr++ = (unsigned char)second;
	*cptr++ = (unsigned char)frame;
	*cptr++ = (unsigned char)subframe;
	if (list)
		cptr += sprintf ((char *)cptr, "%ld", list);
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void zeroListClock (long list)	{ //zero list clock

	cptr = &MSCStr[5];
	*cptr++ = 0x17;
	if (list)
		cptr += sprintf ((char *)cptr, "%ld", list);
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void setMTCchaseOn (long list)	{ //turn MTC chase on

	cptr = &MSCStr[5];
	*cptr++ = 0x19;
	if (list)
		cptr += sprintf ((char *)cptr, "%ld", list);
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void setMTCchaseOff (long list)	{ //turn MTC chase off

	cptr = &MSCStr[5];
	*cptr++ = 0x1A;
	if (list)
		cptr += sprintf ((char *)cptr, "%ld", list);
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void setMTCsourceInternal (long list) {	//make internal MTC generator source of MTC for list

	cptr = &MSCStr[5];
	*cptr++ = 0x06;
	*cptr++ = 0;
	*cptr++ = 0x14;
	*cptr++ = (unsigned char)list;
	*cptr++ = 1;
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void setMTCsourceExternal (long list) {	//make external MTC source of MTC for list

	cptr = &MSCStr[5];
	*cptr++ = 0x06;
	*cptr++ = 0;
	*cptr++ = 0x14;
	*cptr++ = (unsigned char)list;
	*cptr++ = 0;
	*cptr++ = 0xf7;
	sendCommand (MSCStr, cptr - MSCStr);
}
/*************************************************************************************/
void setMidiId (long id) {	//set MIDI ID (writes settings to disk)

	cptr = &sysex11[7];
	*cptr++ = 0;
	*cptr++ = (unsigned char)id;
	*cptr++ = 1;
	sendCommand (sysex11, 11);
	sendCommand (saveSettings, 11);
}
/*************************************************************************************/
void setMidiEcho (long midi, long show, long stopNotify) {	//set MIDI echo flags

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0x18;
	*cptr++ = (unsigned char)midi;
	*cptr++ = ((show&1) << 1) + ((stopNotify&1) << 2);
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
void setDefaultShow (long show)	 { //sets the default show (writes settings to disk)

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0x0E;
	*cptr++ = (unsigned char)show;
	*cptr++ = 0;
	sendCommand (MSC11, 11);
	sendCommand (saveSettings, 11);
}
/*************************************************************************************/
// sets one band of input EQ
//freq = frequency*100, bw = bandwidth*10, fg = gain*10,
//type=1:lowpass, 2:bandpass, 3:highpass
long setInputEQ (long ch, long band, long freq, long bw, long fg, long type) {

	return setEQ (translateChannel(ch-1), band-1, freq, bw, fg, type);
}
/*************************************************************************************/
// sets one band of output EQ
//freq = frequency*100, bw = bandwidth*10, fg = gain*10,
//type=1:lowpass, 2:bandpass, 3:highpass
long setOutputEQ (long ch, long band, long freq, long bw, long fg, long type) {

	return setEQ (translateChannel(ch+63), band-1, freq, bw, fg, type);
}
/*************************************************************************************/
// sets one band of EQ (input/output)
long setEQ (long ch, long band, long freq, long bw, long fg, long type) {

	double w = freq/100;
	double q = bw/10;
	double g = fg/10;
	long DA2, DA1, DA0, DB1, DB0;			// EQ coefficients converted to DSP values
	long result;

	result = DesignFilter (type, w, q, g, &DA2, &DA1, &DA0, &DB1, &DB0);
	if (result) {
		cptr = &setCoeffCmd[8];
		*cptr++ = (unsigned char)ch;
		*cptr++ = (unsigned char)band;
		*cptr++ = DA2 & 0x7F;
		*cptr++ = (DA2 >> 7) & 0x7F;
		*cptr++ = (DA2 >> 14) & 0x7F;
		*cptr++ = (DA2 >> 21) & 0x7F;
		*cptr++ = DA1 & 0x7F;
		*cptr++ = (DA1 >> 7) & 0x7F;
		*cptr++ = (DA1 >> 14) & 0x7F;
		*cptr++ = (DA1 >> 21) & 0x7F;
		*cptr++ = DA0 & 0x7F;
		*cptr++ = (DA0 >> 7) & 0x7F;
		*cptr++ = (DA0 >> 14) & 0x7F;
		*cptr++ = (DA0 >> 21) & 0x7F;
		*cptr++ = DB1 & 0x7F;
		*cptr++ = (DB1 >> 7) & 0x7F;
		*cptr++ = (DB1 >> 14) & 0x7F;
		*cptr++ = (DB1 >> 21) & 0x7F;
		*cptr++ = DB0 & 0x7F;
		*cptr++ = (DB0 >> 7) & 0x7F;
		*cptr++ = (DB0 >> 14) & 0x7F;
		*cptr++ = (DB0 >> 21) & 0x7F;
		*cptr++ = (unsigned char)type;				//set filter parameters
		*cptr++ = freq & 0x7F;
		*cptr++ = (freq >> 7) & 0x7F;
		*cptr++ = (freq >> 14) & 0x7F;
		*cptr++ = bw & 0x7F;
		*cptr++ = (bw >> 7) & 0x7F;
		*cptr++ = fg & 0x7F;
		*cptr++ = (fg >> 7) & 0x7F;
		sendCommand (setCoeffCmd, 39);
	}
	return result;
}
/*************************************************************************************/
void setInputEQbands (long ch, long bands) { //sets number of EQ bands on input channel

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0x0C;
	*cptr++ = (unsigned char)translateChannel (ch-1) ;
	*cptr++ = (unsigned char)bands;
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
void setOutputEQbands (long ch, long bands) { //sets number of EQ bands on output channel

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0x0C;
	*cptr++ = (unsigned char)translateChannel (ch+63);
	*cptr++ =(unsigned char) bands;
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
//for test purposes only. Clears first sector on disk. Forces the same behavior
//as installing a new drive - AB64 will report no firmware on disk
void clearFirmware (void) {

	sysex9[7] = 0x54;
	sendCommand (sysex9, 9);
}
/*************************************************************************************/
void defragmentDisk (void) { //start a disk defragment (currently unimplemented)
//	sysex9[7] = 0x0A;
//	sendCommand (sysex9, 9);
}
/*************************************************************************************/
void setSubMasterGain (long gain)  { //sets submaster gain
//    00: 0.75 dB change every four submaster amplitude steps
//    01: 0.75 dB change every two submaster amplitude steps
//    02: 0.75 dB change every submaster amplitude step
//    03: 1.50 dB change every submaster amplitude step

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0x11;
	*cptr++ = (unsigned char)gain;
	*cptr++ = 0;
	sendCommand (MSC11, 11);
	sendCommand (saveSettings, 11);
}
/*************************************************************************************/
void setSubMasterZero (long zero) { //sets submaster zero (MIDI level 0..127)

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0x12;
	*cptr++ = (unsigned char)zero;
	*cptr++ = 0;
	sendCommand (MSC11, 11);
	sendCommand (saveSettings, 11);
}
/*************************************************************************************/
//sets submaster level (submaster 1..32, MIDI level 0..127)
void setSubMasterLevel (long submaster, long level)  {

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 6;
	*cptr++ = submaster-1;
	*cptr++ = (unsigned char)level;
	sendImmediate = true;
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
//assigns input channels to a submaster (submaster 1..32)
//assign string is comma delimited, can contain ranges (i.e. "1-4, 10, 12-15")
void assignInputsToSubMaster (char *assignString, long submaster)  {

	char *tptr;
	char *dptr;
	long fromCh;
	long toCh;
	long i;

	clearSubmasterAssigns (submaster, 0);
	tptr = strtok (assignString, ",");
	while (tptr) {
		dptr = strchr (tptr, '-');
		if (dptr) {  //range found
			i = sscanf (tptr, "%ld-%ld", &fromCh, &toCh);
			if (i == 2) {
				if (fromCh > 64 || toCh > 64) return;
				for (i = fromCh; i <= toCh; i++)
					submasterAssignInput (i, submaster);
				}
			}
		else {
			i = sscanf (tptr, "%ld", &fromCh);
			if (i == 1) {
				if (fromCh > 64) return;
				submasterAssignInput (fromCh, submaster);
				}
			}
		tptr = strtok (0, ",");
		}
	sendCommand (saveSettings, 11);
}
/*************************************************************************************/
//assigns output channels to a submaster (submaster 1..32)
//assign string is comma delimited, can contain ranges (i.e. "1-4, 10, 12-15")
void assignOutputsToSubMaster (char *assignString, long submaster)  {

	char *tptr;
	char *dptr;
	long fromCh;
	long toCh;
	long i;

	clearSubmasterAssigns (submaster, 1);
	tptr = strtok (assignString, ",");
	while (tptr) {
		dptr = strchr (tptr, '-');
		if (dptr) {  //range found
			i = sscanf (tptr, "%ld-%ld", &fromCh, &toCh);
			if (i == 2) {
				if (fromCh > 64 || toCh > 64) return;
				for (i = fromCh; i <= toCh; i++)
					submasterAssignOutput (i, submaster);
				}
			}
		else {
			i = sscanf (tptr, "%ld", &fromCh);
			if (i == 1) {
				if (fromCh > 64) return;
				submasterAssignOutput (fromCh, submaster);
				}
			}
		tptr = strtok (0, ",");
		}
	sendCommand (saveSettings, 11);
}
/*************************************************************************************/
//      00H-1FH for inputs 1 to 32, 20H-3FH for outputs 1 to 32 (AB64 only)
//      40H-5FH for inputs 33 to 64, 60H-7FH for outputs 33 to 64 (AB64 only)
void submasterAssignInput (long input, long submaster) {

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 6;
	*cptr++ = submaster-1;
	*cptr++ = (unsigned char)translateChannel (input-1) ;
	*cptr++ = 1;
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
void submasterAssignOutput (long output, long submaster) {

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 6;
	*cptr++ = submaster-1;
	*cptr++ = (unsigned char)translateChannel (output+63);
	*cptr++ = 1;
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
//assigns a submaster to a MIDI controller
// (submaster 1..32, MIDI controller 0..127)
void assignSubMasterToController (long submaster, long controller)  {

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 7;
	*cptr++ = (unsigned char)controller;
	*cptr++ = submaster-1;
	*cptr++ = 1;
	sendCommand (MSC11, 11);
	sendCommand (saveSettings, 11);
}
/*************************************************************************************/
//deassigns a submaster from a MIDI controller
// (submaster 1..32, MIDI controller 0..127)
void deassignSubMasterFromController (long submaster, long controller)  {

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 7;
	*cptr++ = (unsigned char)controller;
	*cptr++ = submaster-1;
	*cptr++ = 0;
	sendCommand (MSC11, 11);
	sendCommand (saveSettings, 11);
}
/*************************************************************************************/
//clears input or output channel assignments from a submaster
void clearSubmasterAssigns (long submaster, long IO)  {

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0x20;
	*cptr++ = submaster-1;
	*cptr++ = (unsigned char)IO;
	sendCommand (MSC11, 11);
//	sendCommand (saveSettings, 11);
}
/*************************************************************************************/
//clears all submasters and controller assigns (all submasters, all controllers)
void clearAssigns (void)  {

	cptr = &sysex12[7];
	*cptr++ = 0x0C;
	*cptr++ = 0x51;
	*cptr++ = 0x78;
	*cptr++ = 0x33;
	sendCommand (sysex12, 12);
	sendCommand (saveSettings, 11);
}
/*************************************************************************************/
//clears all settings: MIDI ID, submasters, gain tables, etc.
void clearSettings (void)  {

	cptr = &sysex12[7];
	*cptr++ = 0x0C;
	*cptr++ = 0x76;
	*cptr++ = 0x49;
	*cptr++ = 0x09;
	sendCommand (sysex12, 12);
	sendCommand (saveSettings, 11);
}
//*************************************************************************************
void clearDirectory (void) {

	sendCommand (clearDrive, 12);
}
//*************************************************************************************
void clearMessageCounts (void) {

	sendCommand (clearMsgCounts, 12);
}

/*************************************************************************************/
void setMIDIChannel (int chan)	{ //set channel number for MIDI (Toolset only)

	midiChannel = chan;
}
/*************************************************************************************/
void setRealTimeClock (void)	{ //set AudioBox real-time clock

time_t dlDateTime;
struct tm *tmptr;
struct ab64RealTime rtc;

	time (&dlDateTime);	//get current time
	tmptr = localtime (&dlDateTime); //convert to local time
	memset (&rtc, 0, sizeof (rtc));		//zero the structure
	rtc.code = 0x83;
	rtc.length = htons (sizeof (rtc));
	rtc.validFlags = htons (0x00FB);
	rtc.year = bcd (tmptr->tm_year - 100);
	rtc.month = bcd (tmptr->tm_mon + 1);
	rtc.day = bcd (tmptr->tm_mday);
	rtc.weekday = bcd (tmptr->tm_wday);
	rtc.hour = bcd (tmptr->tm_hour);
	rtc.minute = bcd (tmptr->tm_min);
	rtc.second = bcd (tmptr->tm_sec);
	sendto (dataSock, (const char*)&rtc, sizeof (rtc), 0, ENadrs, ENadrsSize);
}
/*************************************************************************************/
char bcd (long n) {	return (((n/10) << 4) + (n % 10)); }
/*************************************************************************************/
void setCobraNet (short *bundleArray)	{ //set CobraNet parameters

	short assign;
	cptr = &sysex57[7];
	*cptr++ = 0x53;
	long loopCount = 16;
	while (loopCount--) {
		assign = *bundleArray++;
		*cptr++ = assign & 0x7F;
		*cptr++ = (assign >> 7) & 0x7F;
		*cptr++ = (assign >> 14) & 0x7F;
		}
	sendCommand (sysex57, 57);
}
/*************************************************************************************/
void setSendMIDI (long send) { txMIDI = send; }
/*************************************************************************************/
void setGPI (long mode, long lockout) {

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0x09;
	*cptr++ = (unsigned char)mode;
	*cptr++ = lockout & 0x7F;
	*cptr++ = (lockout >> 7) & 0x7F;
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
void setGPO (long output, long state) {

	cptr = &MSC11[5];
	*cptr++ = 6;
	*cptr++ = 0;
	*cptr++ = 0x1F;
	*cptr++ = (unsigned char)output;
	*cptr++ = (unsigned char)state;
	sendCommand (MSC11, 11);
}
/*************************************************************************************/
// Sends command string
void sendCommand (unsigned char* hexStr, long cmdLen) {

long bytecount;

#if HFI
	if (txMIDI) {
		if (midiChannel > 0) {
			bytecount = write (midiChannel, hexStr, cmdLen);
			Delay (1, &endTick);
		}
		return;
		}
#endif

	//put in header
	*sendPacketBufPtr++ = 0x80;
	*sendPacketBufPtr++ = 0;
	*sendPacketBufPtr++ = 0;
	//put in command length, including header
	if (cmdLen & 1)
		*sendPacketBufPtr++ = cmdLen + 5;
	else
		*sendPacketBufPtr++ = cmdLen + 4;
	memcpy (sendPacketBufPtr, hexStr, cmdLen);
	sendPacketBufPtr += cmdLen;
	if (cmdLen & 1)
		*sendPacketBufPtr++ = 0;		//add padding byte
	bytecount = sendPacketBufPtr - sendPacketBuf;
	if (bytecount > 1350 || sendImmediate) {
		sendto (dataSock, (const char*)&sendPacketBuf[0], bytecount, 0, ENadrs, ENadrsSize);     
		sendPacketBufPtr = sendPacketBuf;
		sendImmediate = false;
		}
#if ABSM
	//send message to history controller
	postCommand ((char *)hexStr, cmdLen);
#endif
}
/************************************************************************************
	Filter coefficient calculations, based on
		"Cookbook formulae for audio EQ biquad filter coefficients"
			by Robert Bristow-Johnson
		inType: one of eFTypeLowShelf, eFTypePeaking, eFTypeHighShelf,
			eFTypeLowPass, eFTypeHighPass, eFTypeNotch, eFTypeBandpass
		inHz: centre frequency in Hz
		inBW: bandwidth as % of center freq
		inGain: gain in dB
	5 output coefficients: outA2, outA1, outA0, outB1, outB0
*************************************************************************************/
long DesignFilter (long inType, double inHz, double inBW, double inGain,
			long *outA2, long *outA1, long *outA0, long *outB1, long *outB0) {

double Fs = 48000.0;							// sample rate in Herz
double LN2 = log (2);							// natural log of 2
double mA2,mA1,mA0,mB1,mB0;			// coefficients using AB variables
double b0, b1, b2, a0, a1, a2;				// coefficients using rbj variables

	//convert bandwidth in percent of centre frequency to bandwidth in octaves
	double BW = log ((inBW/100.0) + 1.0) / LN2;

	//calculate intermediate variables
	double A = pow (10.0, inGain / 40.0);
	double w0 = TWOPI * inHz / Fs;
	double cosw0 = cos (w0);
	double sinw0 = sin (w0);
	double alpha = sinw0 * sinh ((LN2/2) * BW * (w0 / sinw0));
	double beta = 2 * sqrt(A) * alpha;
	double gplus = (A + 1) * cosw0;
	double gminus = (A - 1) * cosw0;

	//calculate coefficients based on filter type
	switch (inType) {
		case eFTypeLowShelf:						//was called low pass
			b0 = A * ( (A+1) - gminus + beta);
			b1 = 2 * A * ( (A-1) - gplus);
			b2 = A * ( (A+1) - gminus - beta);
			a0 = (A+1) + gminus + beta;
			a1 = -2 * ( (A-1) + gplus);
			a2 = (A+1) + gminus - beta;
			break;
		case eFTypePeaking:						//was called bandpass
			b0 = 1 + alpha * A;
			b1 = -2 * cosw0;
			b2 = 1 - alpha * A;
			a0 = 1 + alpha / A;
			a1 = -2*cosw0;
			a2 = 1 - alpha / A;
			break;
		case eFTypeHighShelf:						//was called highpass
			b0 = A * ( (A+1) + gminus + beta);
			b1 = -2 * A * ( (A-1) + gplus);
			b2 = A * ( (A+1) + gminus - beta);
			a0 = (A+1) - gminus + beta;
			a1 = 2 * ( (A-1) - gplus);
			a2 = (A+1) - gminus - beta;
			break;
		case eFTypeLowPass:						//classic crossover LPF
			b0 = (1 - cosw0) / 2;
			b1 = 1 - cosw0;
			b2 = b0;
			a0 = 1 + alpha;
			a1 = -2*cosw0;
			a2 = 1 - alpha;
			break;
		case eFTypeHighPass:						//classic crossover HPF
			b0 = (1 + cosw0) / 2;
			b1 = -(1 + cosw0);
			b2 = b0;
			a0 = 1 + alpha;
			a1 = -2 * cosw0;
			a2 = 1 - alpha;
			break;
		case eFTypeBandpass:						//classic bandpass
			b0 = sinw0 / 2;
			b1 = 0;
			b2 = -b0;
			a0 = 1 + alpha;
			a1 = -2*cosw0;
			a2 = 1 - alpha;
			break;
		case eFTypeNotch:							//for eliminating a specific frequency
			b0 = 1;
			b1 = -2 * cosw0;
			b2 = 1;
			a0 = 1 + alpha;
			a1 = -2*cosw0;
			a2 = 1 - alpha;
			break;
		default:
			return false;
		}	//end switch

	//convert rbj variables into AB variables
	mA2 = b0 / a0;
	mA1 = b1 / a0;
	mA0 = b2 / a0;
	mB1 = a1 / a0;
	mB0 = a2 / a0;

	//test for valid coefficient range
	if (mA2 > 2.0 || mA2 < -2.0
	||  mA1 > 2.0 || mA1 < -2.0
	||  mA0 > 2.0 || mA0 < -2.0
	||  mB1 > 2.0 || mB1 < -2.0
	||  mB0 > 2.0 || mB0 < -2.0 )
		return (false);
	else {
		//write output coefficients
		*outA2 = (long)(DSP1 * mA2);  
		*outA1 = (long)(DSP1 * mA1);
		*outA0 = (long)(DSP1 * mA0);
		*outB1 = (long)(DSP1 * mB1);
		*outB0 = (long)(DSP1 * mB0);
		return (true);
		}
}
//*************************************************************************************/
void FrequencyResponse (double theta, double inA2, double inA1, double inA0,
					double inB1, double inB0, double *outMag ) {

	Complex z, z2, znum, zdenom;
	Complex a2, a1, a0, b1, b0;
	z.real = cos(theta);
	z.imag = sin(theta);
	z2 = cmultiply(z,z);

	a2.imag = 0;
	a1.imag = 0;
	a0.imag = 0;
	b1.imag = 0;
	b0.imag = 0;

	a2.real = inA2;
	a1.real = inA1;
	a0.real = inA0;
	b1.real = inB1;
	b0.real = inB0;

		// calculate the complex numerator
	znum = cmultiply (a2, z2);				// a[2] * z2
	znum = cadd (znum, cmultiply(a1, z));	// + a[1] * z
	znum = cadd (znum, a0);					// + a0

		// calculate the complex denominator
	zdenom = z2;							// z2
	zdenom = cadd (zdenom, cmultiply(b1,z));	// + b[1] * z
	zdenom = cadd (zdenom, b0);				// + b[0]

	*outMag = cABS (cdiv(znum, zdenom));		// magnitude = abs(num/denom)
}
//*************************************************************************************
Complex cmultiply (Complex a, Complex b)  {

Complex result;
	result.real = a.real * b.real - a.imag * b.imag;
	result.imag = a.real * b.imag + a.imag * b.real;
	return result;
}
//*************************************************************************************
Complex cdiv (Complex num, Complex denom)  {

Complex result;
double d = denom.real * denom.real + denom.imag * denom.imag;
	result.real = 0;
	result.imag = 0;

	if (d != 0.0) {
	result.real = (num.real * denom.real + num.imag * denom.imag) / d;
	result.imag = (num.imag * denom.real - num.real * denom.imag) / d;
	}
	return result;
}
//*************************************************************************************
Complex cadd (Complex a, Complex b) {

	Complex result;

	result.real = a.real + b.real;
	result.imag = a.imag + b.imag;
	return result;
}
//*************************************************************************************
double cABS (Complex a)  {

	double temp;
	double result = 0;

	temp = a.real * a.real + a.imag * a.imag;
	if(temp >= 0.0)
		result = sqrt (temp);
	return result;
}
/*************************************************************************************/

// uses unix call to implement a timed process wait
void millisleep (long ms) {
#ifdef _FKN_WINDOWS 
	Sleep(ms);
#else
struct timespec tsp;
struct timespec rtsp;

	tsp.tv_sec = 0;
	tsp.tv_nsec = ms * 1000 * 1000;
	nanosleep (&tsp, &rtsp);

#endif
}
/*************************************************************************************/
//prints errString to standard error preceded by a timestamp
void logEvent (char *errString) {
#ifdef _FKN_WINDOWS 
	//timestamp formatted date and time
	char timstr[256];
	char datestr[256];
	char dateformat[] = "d'-'MMM'-'yyyy";
	int siz = GetTimeFormat(NULL, NULL,NULL, NULL,timstr,255);  
	siz = GetDateFormat(NULL, NULL,NULL, dateformat, datestr,255); 	
	fprintf (stderr, "%s %s, %s\n", datestr, timstr, errString);
#else
	//timestamp (seconds, microseconds)
	struct timeval tp;
	gettimeofday(&tp, 0);
	fprintf (stderr, "%d %d %s\n", tp.tv_sec, tp.tv_usec, errString);
#endif
}
/*************************************************************************************/
void logError (char *errString) {
	logEvent (errString);
	errFlag = true;
	strcpy (libError.errorString, errString);
}
/*************************************************************************************/
// rounds a size longword up to the next even page (512 bytes)
void roundUpToPage (long *x) {
	*x = ((*x + 511) / 512) * 512;
}
