/*
 *  ab64if.h
 *
 *  Copyright 2006 Harmonic Functions, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY HARMONIC FUNCTIONS INC. "AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL HARMONIC FUNCTIONS INC. BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *  Created by Tim Bartoo on 06/01/30.
 */
 
//The following is a list of all AB64 response codes returned via the receive 
//message function. These codes occupy the first byte of all incoming AB64 
//message data structures. These data structures are defined in this file and 
//described in the ethernet section of the command set document on the HFI 
//web site (www.hfi.com)
// Code					See Command
// rcAllDataComplete	SET AUTO UPDATE EN
// rcCobranet			GET COBRANET EN
// rcDirectory			GET DIRECTORIES EN
// rcDirUpdateCmpl	GET DIRECTORIES EN
// rcDiskInfo				GET DISK INFO EN
// rcDiskState			GET DISK STATE EN
// rcEQ					GET EQ EN
// rcFileData				GET FILEDATA EN
// rcInquiry				INQUIRY EN
// rcIODelay				GET IO DELAY EN
// rcIOLevels				GET IO LEVELS EN
// rcLCD					GET LCD EN
// rcMuteSolo			GET MUTE/SOLO EN
// rcPlayback			GET PLAYBACK EN
// rcPlbkPoints			GET PLAYBACK POINTS EN
// rcRealTime			GET TIME EN
// rcSelChannels		GET SELECTION CHANNELS EN
// rcShow				GET SHOW EN
// rcStats					GET STATS EN
// rcSubMaster			GET SUBMASTER EN
// rcTooManyHosts		SET AUTO UPDATE EN
// rcVU					GET VU EN
// rcXptDelay			GET XPT DELAY EN
// rcXptLevels			GET XPT LEVELS EN
//  The following response codes originate in ab64if.c itself, rather than the AB64
// rcAB64IFerror			returns error string describing AB64 Interface error
// rcTransferProgress	returns double 0..1 indicating file transfer completion
// rcFromTransferCmpl	indicates transfer from AB64 completed successfully
// rcFWTransferCmpl	indicates firmware transfer completed successfully
 #define rcAllDataComplete 0x7A
 #define rcCobranet 0x67
 #define rcDirectory 0x7B
 #define rcDirUpdateCmpl 0x89
 #define rcDiskInfo 0x65
 #define rcDiskState 0x77
 #define rcEQ 0x78
 #define rcFileData 0x82
 #define rcInquiry 0x6E
 #define rcIODelay 0x69
 #define rcIOLevels 0x68
 #define rcLCD 0x7C
 #define rcMuteSolo 0x75
 #define rcPlayback 0x73
 #define rcPlbkPoints 0x6C
 #define rcRealTime 0x7D
 #define rcSelChannels 0x66
 #define rcShow 0x74
 #define rcStats 0x6F
 #define rcSubMaster 0x76
 #define rcTooManyHosts 0x85
 #define rcVU 0x70
 #define rcXptDelay 0x72
 #define rcXptLevels 0x71
 #define rcAB64IFerror 0xA0
 #define rcTransferProgress 0xA1
 #define rcFWTransferCmpl 0xA2
 #define rcFromTransferCmpl 0xA3

#define samplesPerSecond 48000.00
#define secondsPerMinute 60
#define minutesPerHour 60
#define bytesPerBlock 512
#define blocksPerSecond (samplesPerSecond*2/bytesPerBlock)
#define blocksPerMinute (blocksPerSecond*secondsPerMinute)
#define blocksPerHour (blocksPerMinute*minutesPerHour)
//highest valid selection number
#define hvseln 8063
//highest valid show number
#define hvshn 127
//maximum number of open lists
#define runlistsAllocated 8
//maximum size of v4 path structure in bytes
#define maxv4PathSize 32768
//*************************************************************************************
struct ab64MessageHeader {
	unsigned char code;
	char unused;
	short length;			//message length including this 4 byte header
	};
typedef struct ab64MessageHeader* ab64MsgPtr;
//*************************************************************************************
struct ab64LibError {
	unsigned char code;	//rcAB64IFerror
	char unused;
	short length;	
	char errorString[128];		//C string describing error
	};
//*************************************************************************************
struct ab64TransferProgress {
	unsigned char code;	//rcTransferProgress
	char unused;
	short length;	
	double progress;		//a number between 0 and 1 indicating transfer progress
	};
typedef struct ab64TransferProgress* transferProgressPtr;
//*************************************************************************************
 struct ab64Inquiry {
	unsigned char code;	//rcInquiry
	unsigned char code2;	//0xA7
	short length;
	long IPaddress;			//IP address
	short inchannels;			//number of input channels (0, 16, 32, 48, or 64) (AB64 only)
	short outchannels;		//number of output channels (0, 16, 32, 48, or 64) (AB64 only)
	char mfg[8];
	char product[16];
	char version[4];
	char date[8];
	char time[8];
	char serialnum[8];
	char unitName[32];
	};
typedef struct ab64Inquiry* inquiryPtr;
//*************************************************************************************
struct ab64SelectionChannels {
	unsigned char code;	//rcSelChannels
	char unused;
	short length;
	long segment; 
	char channels[1024];
	};
typedef struct ab64SelectionChannels* selChanPtr;
//*************************************************************************************
struct ab64IOLevels {
	unsigned char code;	//rcIOLevels
	char unused;
	short length;
	char analogSrcLev[64];
	char digitalSrcLev[64];
	char playbackSrcLev[64];
	char inputLev[64];
	char outputLev[64];
	};
typedef struct ab64IOLevels* IOLevelsPtr;
//*************************************************************************************
struct ab64IODelay {
	unsigned char code;	//rcIODelay
	char unused;
	short length;
	unsigned char inputEnable[64];
	unsigned char outputEnable[64];
	long inputDelay[64];
	long outputDelay[64];
	};
typedef struct ab64IODelay* IODelayPtr;
//*************************************************************************************
struct ab64PlbkPoints {
	unsigned char code;	//rcPlbkPoints
	char unused;
	short length;
	struct {
		long stopSelection;   //selection number containing stop point
		long stopLocation;    //stop point location in blocks from start of selection, 0 means none
		long resumeSelection; //selection number containing resume point
		long resumeLocation;  //resume point location in blocks from start of selection, 0 means none
		} plbkrecord[64];
	};
typedef struct ab64PlbkPoints* plbkPointsPtr;
//*************************************************************************************
struct ab64Stats {
	unsigned char code;		//rcStats
	char unused;
	short length;
	long freeBuffers;			//number of entries on the free queue
	long lowWaterMark;  		//free queue low water mark
	long timerCount;			//number of entries on the timer queue
	long workCount;			//number of entries on the work queue
	long MIDI0TXCount;		//number of entries on the MIDI 0 transmit queue
	long MIDI1TXCount;		//number of entries on the MIDI 1 transmit queue
	long currentTime;			//current millisecond count
	long cpuUtilization;			//CPU utilization percent
	long enetMsgs;				//messages received via ethernet
	long enetOKmsgs;			//valid messages recevied via ethernet
	long enetBADmsgs;			//invalid messages recevied via ethernet
	long MIDI0rxMsgs;			//messages received via MIDI port 0
	long MIDI0OKmsgs;		//valid messages recevied via MIDI port 0
	long MIDI0BADmsgs;		//invalid messages recevied via MIDI port 0
	long MIDI1rxMsgs;			//messages received via MIDI port 1
	long MIDI1OKmsgs;		//valid messages recevied via MIDI port 1
	long MIDI1BADmsgs;		//invalid messages recevied via MIDI port 1
	long MIDI0txMsgs;			//messages transmitted via MIDI port 0
	long MIDI1txMsgs;			//messages transmitted via MIDI port 1
	long temperature;			//internal unit temperature
	char reserved[28];
	};
typedef struct ab64Stats* statsPtr;
//*************************************************************************************
struct ab64VU {
	unsigned char code;		//rcVU
	char unused;
	short length;
	char inputVU[64];
	char outputVU[64];
	};
typedef struct ab64VU* vuPtr;
//*************************************************************************************
struct ab64XptLevels {
	unsigned char code;		//rcXptLevels
	char unused;
	short length;
	long startOutChannel;
	char xptLev[16][64]; //[output][input]
	};
typedef struct ab64XptLevels* xptLevelsPtr;
//*************************************************************************************
struct ab64XptDelay {
	unsigned char code;		//rcXptDelay
	char unused;
	short length;
	long startOutChannel;
	long xptDelay[4][64]; //[output][input]
	};
typedef struct ab64XptDelay* xptDelayPtr;
//*************************************************************************************
struct ab64PlbkChannel {
	char unused;
	char recording;			//nonzero: channel is being recorded
	char stopped;           //nonzero: channel is stopped
	char disabled;          //nonzero: channel is disabled
	long selectionNumber;   //selection number loaded on channel
	long playbackLocation;  //playback location in blocks from start of selection
	long emptyBlocks;       //empty blocks in audio buffer
	};
struct ab64Playback {
	unsigned char code;		//rcPlayback
	char unused;
	short length;
	struct ab64PlbkChannel pch[64];
	};
typedef struct ab64Playback* plbkPtr;
//*************************************************************************************
struct ablistdata {
	char dataValid;				//structure holds valid data, list and path are open
	char lastCueRun;				//non-zero indicates last cue in path has been fired
	unsigned short listNumber; //list number
	unsigned short listFlags;  //list flags.
	unsigned short pathNumber; //path number
	unsigned short nCues;		//number of cues in path
	unsigned short nLiveCues; //number of live cues in path
	unsigned long listClock; //list clock value (milliseconds)
	unsigned long seqClock;		//sequence clock value (milliseconds)
	unsigned short lastCuenumber; //major number for next cue
	unsigned char lastSub1; //1st subsection number for next cue
	unsigned char lastSub2; //2nd subsection number for next cue
	unsigned char lastSub3; //3rd subsection number for next cue
	unsigned char lastSub4; //4th subsection number for next cue
	unsigned short nextCuenumber; //major number for next cue
	unsigned char nextSub1; //1st subsection number for next cue
	unsigned char nextSub2; //2nd subsection number for next cue
	unsigned char nextSub3; //3rd subsection number for next cue
	unsigned char nextSub4; //4th subsection number for next cue
};
//Show Runtime Data returned by the GETSHOW command
struct ab64ShowData {
	unsigned char code;		//rcShow
	char unused;
	short length;
	unsigned short showNumber; //show number (0 = none open)
	char showName[32];			//show name
	unsigned short numPaths; //number of paths
	unsigned short numLists; //number of lists
	char MIDIecho;      			//boolean, 0: off, non-zero: on
	char MIDIloopback;   		//boolean, 0: off, non-zero: on
	long MTCgenerator;         //current value of the MTC generator in ms
	char MTCgenFrame;          //MTC frames/sec (24, 25, 30, 0 = disabled)
	char MTCgenRunning;        //MTC generator running (non-zero: running)
	char reserved[30];
	struct ablistdata listdata[runlistsAllocated];
	};
typedef struct ab64ShowData* showDataPtr;
//*************************************************************************************
struct ab64MuteSolo {
	unsigned char code;		//rcMuteSolo
	char unused;
	short length;
	char inputMuteSolo[64]; //input channel mute/solo settings
	char outputMuteSolo[64]; //output channel mute/solo settings
	};
typedef struct ab64MuteSolo* muteSoloPtr;
//*************************************************************************************
struct ab64Submaster {
	unsigned char code;		//rcSubMaster
	char unused;
 	short length;
	short submasterGain;  //current submaster gain, see SET SUBMASTER GAIN
	short submasterZero;  //current submaster zero, see SET SUBMASTER ZERO
	char submasterLevels[32]; //current submaster levels (-128..+127)
	char submasterAssigns[32][16]; //current submaster assigns
	char controllerAssigns[128]; //current controller assigns
};
typedef struct ab64Submaster* submasterPtr;
//*************************************************************************************
struct ab64DiskState {
	unsigned char code;	//rcDiskState
	char unused;
	short length;
	char diskEnable;    //playback system is enabled; ready for playback
	char playbackActive;//one or more channels are playing back
	char dirError;      //error occured on last directory verify; one or more files were erased
	char transferActive;//audio or show file transfer is in progress
	char eraseActive;   //audio or show file erasure is in progress
	char defragActive;  //disk defragment is in progress
	char formatActive;  //disk format operation is in progress
	char dirClrActive;  //clear directory is in progress (CLEAR DATA)
	char MotorMixActive;//MotorMix interface has been activated
	char AllOffLock;    //ALLOFF lock
	char transferRate;   //reserved
	char audioMode;		//audio mode
	long diskCapacity;     //total capacity of the disk in blocks
	long nowfreeBlks;      //number of blocks now free on the disk
	long freeBlocks;       //number of blocks free after defragmentation
	long highestSelNum;       //highest selection number in use
	long defragMarker;  //defragment marker (for progress indicator)
	long defragEnd;     //defragment endpoint (for progress indicator)
	long highestShNum;       //highest show number in use
	char GPOstate;			//state of GPO outputs, bit 0: GPO 1, bit 1: GPO 2
	char autoConfIP;     //using auto-configured IP address
	char slaveDisk;      //0: using master/primary disk, 1: using slave/secondary 
	char MIDItestResults;     //boot MIDI loopback test results: bit 0: port 1, bit 1: port 2
	char unused2[24];   //reserved
	char MIDIID;        //MIDI device ID
	char MIDIchannel;   //assigned MIDI channel (FFH: all channels)
	short MIDIgroupIDs; //MIDI group IDs assigned
	char defaultShow;   //default show number
	char allOffActions; //MSC ALL_OFF action flags
	char resetActions1; //MSC RESET action flags, byte 1
	char resetActions2; //MSC RESET action flags, byte 2
	char submasterGain; //current submaster gain setting
	char submasterZero; //current submaster zero setting
	char fixedIP;       //fixed IP flag - set if ABFIXEDIP is in use (returned in INQUIRY)
	char GPIplbkTrigMode; //GPI playback trigger mode
	long GPItrigLockout; //GPI trigger lockout time in milliseconds
	char unused4[32];    //reserved
   };
typedef struct ab64DiskState* diskStatePtr;
//*************************************************************************************
struct ab64EQch {
	long nSections;  //current number of EQ bands
	struct {
		long a2;      //coefficient a2
		long a1;      //coefficient a1
		long b1;      //coefficient b1
		long a0;      //coefficient a0
		long b0;      //coefficient b0
		short filterType; // filter type: 0:none, 1:lowpass, 2:highpass, 3:bandpass
		unsigned short bandwidth;  // bandwidth * 10 (valid range is 10..990 for 1%..99% of centre frequency)
		signed short gain;  //gain in dB * 10 (e.g. 156 for +15.6 dB)
		short pad;          //for longword alignment
		unsigned long frequency; // centre frequency * 100, 2000..2000000 (e.g. 22575 for 225.75 Hz)
		} section[7];
	};
struct ab64EQ {
	unsigned char code;		//rcEQ
	char unused;
 	short length;
	 long IOflag;   //0: input, 1: output
	 long channel;  //channel number: 0..63
	struct ab64EQch eqch;
	};
typedef struct ab64EQ* eqPtr;
//*************************************************************************************
struct ab64Directory {
	unsigned char code;		//rcDirectory
	char unused;
 	short length;
	long dirType; //0: audio, 1: show
	long directorySegment; 
	long blocks[128];
};
typedef struct ab64Directory* directoryPtr;
//*************************************************************************************
struct ab64DirUpdateCmpl {
	unsigned char code;		//rcDirUpdateCmpl
	char unused;
 	short length;
	long dirType; //0: audio, 1: show
	long fileNumber; 
};
typedef struct ab64DirUpdateCmpl* dirUpdateCmplPtr;
//*************************************************************************************
struct ab64LCD {
	unsigned char code;		//rcLCD
	char unused;
 	short length;
	unsigned char screenBits[128][8]; //[bitmap bytes][rows]
};
typedef struct ab64LCD* lcdPtr;
//*************************************************************************************
 struct ab64RealTime {
	unsigned char code;		//rcRealTime
	char unused;
	short length;
	unsigned short validFlags;      //field valid flags
	unsigned char year;             //year, '00' for year 2000 .. '99'
	unsigned char month;            //month, '01' .. '12' 
	unsigned char week;             //week of year, '01' .. '53'
	unsigned char day;              //day of month, '01 .. '31'
	unsigned char weekday;          //day of week, '01 .. 07' (Sunday = 1)
	unsigned char hour;             //hour, '00' .. '23'
	unsigned char minute;           //minute, '00' .. '59'
	unsigned char second;           //second , '00' .. '59'
	short sunrise;  //minutes before/after sunrise -1440..1440 (unused)
	short sunset;   //minutes before/after sunset  -1440..1440 (unused)
	unsigned char daynight;        //night=0, day=1
	unsigned char dst;             //standard=0, daylight savings time=1
	unsigned char reserved[16];
	};
typedef struct ab64RealTime* realTimePtr;
//*************************************************************************************
	struct ab64DiskInfo {
		short reserved1[3];
		short errorNumber;				//AB64 disk error number, if error during disk initialization
		char ready1;						//disk is ready to read and write
		char ready2;						//AB64 internal use
		short resetDiagnosticCode;	//ATA spec, table 23
		unsigned short capacityHi;	//disk capacity (copied from capacityHighWord)
		unsigned short capacityLo;	//disk capacity (copied from capacityLowWord)
		char use28bitAdrs;				//drives uses older 28 bit addressing
		char unused2[15];
		short general;					//structure from this field on is exactly as read from
		short cylinders;					//disk via ATA IDENTIFY DEVICE command
		short specificConfig;			//see ATA spec for field descriptions
		short heads;
		short retired1[2];
		short sectorsPerTrack;
		short reserved2[3];
		char serialNumber[20];
		short retired3[2];
		short obsolete1;
		char firmwareRevision[8];
		char model[40];
		short sectorsPerMultCmd;
		short reserved3;
		short capabilities1;
		short capabilities2;
		short obsolete2[2];
		short validBits;
		short currentCylinders;
		short currentHeads;
		short currentSectorsPerTrack;
		unsigned short capacityLowWord;
		unsigned short capacityHighWord;
		short sectorsPerInterrupt;
		unsigned short addressableSectorsLowWord;
		unsigned short addressableSectorsHighWord;
		short obsolete3;
		short DMAmode;
		short PIOModes;
		short minDMAcycle;
		short recommendDMAcycle;
		short minPIOwithoutFC;
		short minPIOwithFC;
		short reserved4[6];
		short queueDepth;
		short reserved5[4];
		short majorVersion;
		short minorVersion;
		short cmdSetSupportedBits1;
		short cmdSetSupportedBits2;
		short cmdSetSupportedBits3;
		short cmdSetEnabledBits1;
		short cmdSetEnabledBits2;
		short cmdSetEnabledBits3;
		short ultraModes;
		short securityEraseTime;
		short enhancedSecurityEraseTime;
		short advancedPowerManValue;
		short masterPasswordRevCode;
		unsigned short hardwareResetResult;
		short acousticManValue;
		short reserved6[5];
		unsigned short maxLBA48word1;
		unsigned short maxLBA48word2;
		unsigned short maxLBA48word3;
		unsigned short maxLBA48word4;
		short reserved7[152];
		};
typedef struct ab64DiskInfo* diskInfoPtr;
struct ab64DiskInfoMsg {
	unsigned char code;	//rcDiskInfo
	char unused;
	short length;
	struct ab64DiskInfo diskinfo[2];	//master disk info is in diskinfo[0], slave is in diskinfo[1]
	};
typedef struct ab64DiskInfoMsg* diskInfoMsgPtr;
//*************************************************************************************
#define srcPathNameStringLen 200
struct ab64FileData {
	char filename[32]; //short selection/show name string
	long dlDate; //download date/time (seconds since May 30, 1997, 3:00 AM)
	char flags;   //see flag bit definitions below
	char reserved[11];  //reserved for future definition
	long sourceFileSize; //size of the source file
	long modDate; //source file modification date/time (seconds from May 30, 1997, 3:00 AM)
	char sourcePathName[srcPathNameStringLen];  //source file path
	};

struct ab64GetFileData {
	unsigned char code;
	char unused;
	short length;
	long ftype;		//file type, 0=audio, 1=show
	long stfn;			//starting file number
	struct ab64FileData fd[4];
 };
typedef struct ab64GetFileData* getFileDataPtr;
//*************************************************************************************
//*************************************************************************************
// Common show file structures, versions 4 and 5
//*************************************************************************************
//List structure (version 4 & 5):
//Two 16-bit words, the first is the Note Off Mask, the second
//is a flag word. Up to 128 lists can be defined per show, with 
//at most 'runlistsAllocated' open at any time. MTC Sequence 
//Timing determines the source for the timing of the events 
// within a cue (the sequence). Clock type is either stopwatch
// (internal timing) or incoming MTC. Auto cue path sequencing, 
//if on, tells the firmware to automatically load the next defined 
//path (next higher path number) when the last cue of the current 
//path is completed. The Clock Running flag, if on, sets the list clock 
//running as soon as the list is opened. The 'List Opened when 
//Show Opened' flag, if on, indicates that this list will be opened
//automatically when the show is opened  The firmware will scan
//the list array from the beginning, and open the first two lists it
//finds having this flag set.
//   List Flags:
//    Bit
//     0&1:   Sequence Timing: 00: folow list clock, whatever it is
//                                 01: Always stopwatch (internal)
//                                 02: Always MTC
//      2:    Clock type: 0: stopwatch (internal), 1: MTC
//      3:    Auto cue path sequencing, 0: off, 1: on
//      4:    Clock Running  0: off, 1: on
//      5:    List Opened when Show Opened   0: off, 1: on
//      6:    MTC Source  0: external, 1: internal
//      7:    List Clock Slave Mode  0: off, 1: on
//      8:    Real-time List  0: normal list, 1: real-time list
//     all other bits currently undefined
struct ablist {
   unsigned short noteOffMask; //bit number = MIDI channel, bit set = send Note Off on abnormal list stop
   unsigned short flags;       //list flags.
   };
#define ablistsize 4
//*************************************************************************************
//Real Time structure (version 4 & 5):
struct abrealtime {
   char year;				//year, '00' for year 2000 .. '99'
   char month;				//month, '01' .. '12' 
   char week;				//week of year, '01' .. '53'
   char day;				//day of month, '01 .. '31'
   char weekday;			//day of week, '01 .. 07' (Sunday = 1)
   char hour;				//hour, '00' .. '23'
   char minute;			//minute, '00' .. '59'
   char second;			//second , '00' .. '59'
   short sunrise;			//minutes before/after sunrise -1440..1440 (unused)
   short sunset;			//minutes before/after sunset  -1440..1440 (unused)
   char daynight;			//night=0, day=1
   char dst;				//standard=0, daylight savings time=1
   char maskyear;			//mask- year
   char maskmonth;		//mask- month
   char maskweek;			//mask- week of year
   char maskday;			//mask- day of month
   char maskweekday;		//mask- day of week
   char maskhour;			//mask- hour
   char maskminute;		//mask- minute
   char masksecond;		//mask- second
   short masksunrise; //mask- minutes before/after sunrise (unused)
   short masksunset;		//mask- minutes before/after sunset (unused)
   char maskdaynight; //mask- day / night
   char maskdst;			//mask- daylight savings time
   char unused[4];		//reserved, set to zero
   };
typedef struct abrealtime* abrealtimePtr;
#define abrealtimesize 32
//*************************************************************************************
//Description structure  (version 4 & 5):
//Used for comment and file name strings in the show file (ABDS and AFNS chunks)
struct onedescription {
	 unsigned char lengthLo;     // lo-byte of string length
	 unsigned char lengthHi;	    //hi-byte of string length
	 char data[2];	// variable length ascii string (empty strings have no data here)
}; 
struct descHdr {
	long chunkid;
	long chunkSize;
	short listcount;  //# of list descriptions
	short othercountHi;  // # of descriptions for all others (cue and event), upper word
	short othercountLo;  // # of descriptions for all others (cue and event), lower word
	};
typedef struct descHdr* descHdrPtr;
#define descHdrSize 14
// Note re; the descHdr structure: The structure as originally defined had a longword for othercount:
//	struct descHdr {
//		long chunkid;
//		long chunkSize;
//		short listcount;  //# of list descriptions
//		long othercount;  // # of descriptions for all others (cue and event)
//		};
//Compilers in use when the structure was first defined always packed this structure, that is, there was
// no padding between listcount and othercount. Nowadays, most compilers put padding bytes in to
// align longwords, which results in incorrect field references to this structure, which must be kept 
// in its original form to maintain compatibility with existing files which contain this structure.
// There are two solutons. One is to use compiler flag -fpack-struct. This works, but is a detail
// that can easily be lost moving between projects and compilers. It also results in slightly slower code 
// for the entire project.  The other is to define two shorts as is done above, and set the values of each half 
// of the longword separately in the code. And use descHdrSize rather than sizeof (struct descHdr). 
// This removes the compiler dependency.
//*************************************************************************************
//Diffusion chunk ('DIFF' & 'DFUS') structures
// DFUS chunk is structured as follows: a chunk header, followed by 
// 	'numberOfSpkrMaps' structures, each of which is a speakerMap
//	structure followed by 'numberOfSpkrs' speaker structures, then
// 'numberOfDiffusions' structures, each of which is a diffusion
// structure followed by 'numberOfTrajectories' structures, each of which
// is a trajectory structure followed by 'numberOfPoints' 
// soundPoint structures. 
// DIFF is an earlier version. It is similar, but without the diffusion structure. 
struct DIFFchunkHdr {
	long chunkid;
	long chunkSize;
	short numberOfSpkrMaps;		//number of speaker map structures in the chunk
	short numberOfTrajectories;		//number of diffusion structures in the chunk
	};
typedef struct DIFFchunkHdr *DIFFchunkHdrPtr;
#define DIFFchunkHdrSize sizeof (struct DIFFchunkHdr)

struct DFUSchunkHdr {
	long chunkid;
	long chunkSize;
	short numberOfSpkrMaps;		//number of speaker map structures in the chunk
	short numberOfDiffusions;		//number of diffusion structures in the chunk
	char unused[32];					//for later
	};
typedef struct DFUSchunkHdr *DFUSchunkHdrPtr;
#define DFUSchunkHdrSize sizeof (struct DFUSchunkHdr)

struct DFUSspeakerMap {
	char name[32];    //ASCII null-terminated C string
	long numberOfSpkrs;
	long originX;		//float coordinate x 1000
	long originY;		//float coordinate x 1000
	};
typedef struct DFUSspeakerMap *DFUSspkrMapPtr;
#define DFUSspeakerMapSize sizeof (struct DFUSspeakerMap)

struct DFUSspeaker {
	long outputCh;
	long x;		//float coordinate x 1000
	long y;		//float coordinate x 1000
	};
typedef struct DFUSspeaker *DFUSspkrPtr;
#define DFUSspeakerSize 12

struct DFUSdiffusion {						//added going from DIFF to DFUS
	long startEvent;						//sequence number in show of first event in diffusion
	long numberOfEvents;				//number of events in the diffusion
	long numberOfTrajectories;		//number of trajectories in the diffusion
	char unused[32];					//for later
	};
typedef struct DFUSdiffusion *DFUSdiffPtr;
#define DFUSdiffusionSize sizeof (struct DFUSdiffusion)

struct DIFFtrajectory {
	char spkrMapName[32];			//ASCII null-terminated C string
	short inputCh;						//input channel 0..63
	short numberOfPoints;				//number of soundPoint records in this trajectory
	short listNumber;					//diffusion cue list number
	short pathNumber;					//diffusion cue path number
	long cueNumber;					//diffusion cue number
	unsigned long cueSubNumber;	//diffusion cue sub number
	char unused[32];					//for later
	};
typedef struct DIFFtrajectory *DIFFtrajPtr;
#define DIFFtrajectorySize sizeof (struct DIFFtrajectory)

struct DFUStrajectory {
	char spkrMapName[32];			//ASCII null-terminated C string
	short inputCh;						//input channel 0..63
	short numberOfPoints;				//number of soundPoint records in this trajectory
	char repeatCount;					//number of times trajectory repeats in diffusion
	char unused[31];					//for later
	};
typedef struct DFUStrajectory *DFUStrajPtr;
#define DFUStrajectorySize sizeof (struct DFUStrajectory)

struct DFUSsoundPoint {
	long x;								//float coordinate x 1000
	long y;								//float coordinate x 1000
	long diameter;						//float diameter x 1000
	long transitionTime;				//float transitionTime x 1000
	long pauseTime;					//float pauseTime x 1000
	char setInputLevel;					//boolean, set input
	char inputLevel;						//input level 0..127
	char zeroPoint;						//boolean, point generates zero levels
	char unused[17];					//for later
	};
typedef struct DFUSsoundPoint *DFUSsndPntPtr;
#define DFUSsoundPointSize sizeof (struct DFUSsoundPoint)
//*************************************************************************************
// AudioBox Show File Structures, version 4
// Note: all AB1616 and 1616HD models, and AB64 models running firmware
// prior to A439 use the version 4 show file structure.
//*************************************************************************************
//Show File structure:
//A version 4 AudioBox show file is structured as follows:
// a show header
// a list array
// a path index - as many abpathindexv4 structures as there are paths in the file
// a variable number of variable length path structures
// a variable number of variable length optonal chunks
//struct abshow {
//   struct abshowhdrv4 hdr;             //the show header
//   struct ablist list[128];           //list array
//   struct abpathindexv4 pathindex[1];         //path index
//   struct abpathv4[];              //heap of all the path structures
//   };
//*************************************************************************************
//Show Header:
//The version 4 show header contains a file ID for verifying the file type, 
//a version number, a name string, a field for the number of 
//path structures in the file, and some show global flags. 
//If autoClockStart is on, after a RESET or ALL_OFF, a GO on 
//any cue restarts the clock. If MIDIecho is on, all MIDI 
//messages which show up at the MIDI IN port get echoed 
//to the MIDI OUT port. If MIDIloopback is on, the matrix 
//and playback portions of the AudioBox are listening to 
//the output of the show control engine, if off the show 
//control engine is just driving the MIDI OUT port.
struct abshowhdrv4 {
   long fileID;                  //constant 'ABSH' (0x41425348) to identify file 
   short vers;                   //version number for this file structure
   char showname[32];            //show name text string
   short npaths;                 //number of path structures in the file
   unsigned char autoClockStart; //boolean, 0: off, non-zero: on
   unsigned char MIDIecho;       //boolean, 0: off, non-zero: on
   unsigned char MIDIloopback;   //boolean, 0: off, non-zero: on
   unsigned char fps;	    	  //frames per second (24, 25, or 30)
   long optchunkofs;					//offset from start of file to 1st optional chunk
   unsigned long setupPathOffset;     //offset in bytes from start of file to 
                                 //start of setup path structure 
   unsigned char reserved[24];     //reserved for future use
   };
typedef struct abshowhdrv4* abv4shHdrPtr;
#define abshowhdrsizev4 76
//*************************************************************************************
//Path Index version 4:
//There are as many of these are there are path structures in the
//file (npaths). Each entry consists of the path identifier, which 
//is a combination of list number and path number, and the offset
//from the start of the file to the path structure.
struct abpathindexv4 {
   unsigned short listnumber;   //list number for this path
   unsigned short pathnumber;   //path number
   unsigned long pathoffset;    //offset in bytes from start of file to start of path structure 
   };
#define abpathindexsizev4 8
//*************************************************************************************
//Path structure version 4:
//The version 4 path structure consists of a header, a variable number 
//(ncues) of cue structures, followed by a heap of event 
//structures. (Because the event structures are variable-length, 
//it is not actually an array, although array syntax is used below 
//to suggest the structure.) The event structures for a given cue 
//are in a contiguous string, and sorted in order of delay 
//(longer delays after shorter delays).
struct abpathv4 {
	long structID;    //constant 'ABPA' (0x41425041)
	short ncues;      //number of cues in this path
	short evHeapOfs;  //offset from base of path structure to base of event heap
	char desc[64];    //text field
	char unused[32];  //for future needs
//   struct abcuev4[];   //variable length array of cue structures
//   struct abrealtime[];   //variable length array of real time structures
//   struct abeventv4[]; //variable length heap of event structures for this path
   };
typedef struct abpathv4* abpathv4Ptr;
#define abpathsizev4 104
//*************************************************************************************
//Cue structure version 4:
//The event1offset is the offset in bytes from the first event in 
//the event heap to the first event for this cue. The first event in 
//the first cue has event1offset set to zero.
struct abcuev4 {
   unsigned char hdr[12]; // reserved for use by AB (data in file: zeros)
   long cuetime;       // MTC or internal clock time (milliseconds) to GO cue - 0 means manual cue
   long event1offset;  //offset from base of event heap to 1st event in cue
   unsigned short nevents; //number of events in the cue
   unsigned short flags; //cue flags. Defined as of now: bit 0: Zero Clock flag
   unsigned short cuenumber; //major number for this cue
   unsigned char sub1; //1st subsection number for this cue
   unsigned char sub2; //2nd subsection number for this cue
   unsigned char sub3; //3rd subsection number for this cue
   unsigned char sub4; //4th subsection number for this cue
   };
typedef struct abcuev4* abcuev4Ptr;
#define abcuev4size 30
// Note re; the abcuev4 structure:
//Most C compilers generate sizeof (struct abcuev4) == 32 without compiler flag -fpack-struct
//sho file has historically used packed structures, i.e. a cue is 30 bytes, not 32.
//The fields are in the right locations, it is just that 2 extra padding bytes are added at the end if
//	-fpack-struct is not there. Therefore the only place where this shows up is in advancing through
// an array of cue structures by incrementing or decrementing a pointer to an abcuev4 structure.
//*************************************************************************************
//Event structure, version 4:
//A MIDI message with a length, delay value, port number and 
//a header. The port number is for switching the MIDI port if a 
//MIDI switcher is in use. Structure must be word-aligned, 
//which in the event of an even-lengthed MIDI message requires 
//adding a padding byte at the end of the message.
struct abeventv4 {
   unsigned char hdr[10]; // reserved for use by AB (data in file: zeros)
   unsigned short evdelayHi; // delay (milliseconds) from when cue GOes to when event GOs, upper word
   unsigned short evdelayLo; // delay (milliseconds) from when cue GOes to when event GOs, lower word
   unsigned short evsize; //number of bytes in the MIDI message
   unsigned char port;    // MIDI port number
   unsigned char msg[129];   //variable length MIDI message
//   unsigned char pad;     //if necessary to make structure an even size
   };
typedef struct abeventv4* abeventv4Ptr;
#define abeventv4HeaderSize 17
// Note re; the abeventv4 structure: The structure as originally defined had a longword for event delay:
//		struct abeventv4 {
//			unsigned char hdr[10]; // reserved for use by AB (data in file: zeros)
//			unsigned long evdelay; // delay (milliseconds) from when cue GOes to when event GOs
//			...
//Compilers in use when the structure was first defined always packed this structure, that is, there was
// no padding between hdr and evdelay. Nowadays, most compilers put padding bytes in to
// align longwords, which results in incorrect field references to this structure, which must be kept 
// in its original form to maintain compatibility with existing files which contain this structure.
//There are two solutons. One is to use compiler flag -fpack-struct. This works, but is a detail
//that can easily be lost moving between projects and compilers, and results in slightly slower code
// for the entire project. The other is to define two shorts as is done above, and set the values of 
// each half of the longword separately in the code. This removes the compiler dependency.
//*************************************************************************************
//*************************************************************************************
// AudioBox Show File Structures, version 5
//*************************************************************************************
//Show Header, version 5 :
//The show header contains a file ID for verifying the file type, 
//the version number (5), fields for the number of 
//path and list structures in the file, and some show global flags. 
struct abshowhdrv5 {
	long fileID;				//constant 'ABSH' (0x41425348) to identify file 
	short vers;				//0x0005: version number for this file structure
	char unused1[2];		//reserved, zero this field
	long nlists;				//number of lists in the file
	long npaths;				//number of path structures in file (setup not included)
	char autoClockStart;	//boolean, 0: off, non-zero: on
	char MIDIecho;			//flags (see SET MIDI ECHO)
	char MIDIloopback;		//boolean, 0: off, non-zero: on
	char fps;					//frames per second, (24, 25, or 30)
	long optchunkofs;		//offset from start of file to 1st optional chunk
	long setupCueOffset;	//file address (offset) of setup cue structure (multiple of 512)
	long setupCueSize;		//size of setup cue structure (multiple of 512)
	char unused2[480];		//reserved, zero this field
	};
typedef struct abshowhdrv5* abv5shHdrPtr;
#define abshowhdrsizev5 512
//*************************************************************************************
//Path Table (version 5 only):
//The path table immediately follows the show header (at file address 1024). The 
//path table is variable in size and padded with zeros up to an even number of disk 
//blocks (512 bytes). The path table consists primarily of an array of path index 
//structures, one for each path in the file and sorted by list and path. (e.g., all path 
//index structures for list 1 are in path number order and located before all path 
//index structures for list 2, etc.).
struct abpathtablev5 {
	long size;						//size of path table structure, multiple of 512	
	long unused1;				//used by AB64, zero this field
	long structID;				//constant 'PTBL' (0x5054424C) to identify structure
	long unused2;				//reserved, zero this field
    };
typedef struct abpathtablev5* abv5pathTblPtr;
#define abpathtablesizev5 16
//Variable size path index array immediately follows path table header
//There are as many abpathindexv5 structures as there are abpathv5 structures
// in the file (npaths).
struct abpathindexv5 {
	short listnumber;			//list number
	short pathnumber;			//path number
	char openPath;				//flag: open path when show opened (list flags bit 5 + first path in list)
	char unused1[3];			//reserved, zero this field
	long address;				//file address (offset) of path structure (multiple of 512)
	long size;						//size of path structure (multiple of 512)
	};
typedef struct abpathindexv5* abv5pathindexPtr;
#define abpathindexsizev5 16
//*************************************************************************************
//Path structure, version 5:
//The path structure is varaible in size and padded with zeros up to an even number 
//of disk blocks (512 bytes). The path structure contains an array of cue index 
//structures, sorted by cue number. If the path is in a real-time list, an array of 
//realtime structures follows the cue index array, in the same order as the cue index 
//structures (equivalently, a cue index and its corresponding real-time structure both 
//have the same array index).
struct abpathv5 {
	long size;						//size of path structure, multiple of 512
	long unused1;				//used by AB64, zero this field
	long structID;				//constant 'ABPA' (0x41425041)
	long unused2;				//reserved, zero this field
	short listnumber;			//list number
	short pathnumber;			//path number
	long ncues;					//number of cues in this path (must be > 0)
	long listFlags;				//list flags (this field identical for all paths in list)
	long listNoteOffMask;		//list note off mask (this field identical for all paths in list)
	char unused3[32];			//reserved, zero this field	
	};
typedef struct abpathv5* abpathv5Ptr;
#define abpathsizev5 64
//Variable size cue index array immediately follows abpathv5 path header
//There is one abcueindexv5 structure for every cue in the path (ncues)
struct abcueindexv5 {
	long cuenumber;				//major cue number (1..65535)
	unsigned char subnumber1;	//first sub number (0..255)
	unsigned char subnumber2;	//second sub number (0..255)
	unsigned char subnumber3;	//third sub number (0..255)
	unsigned char subnumber4;	//fourth sub number (0..255)
	long address;					//file address of cue structure (multiple of 512)
	long size;							//size of cue structure (multiple of 512)
	};
typedef struct abcueindexv5* abcueindexv5Ptr;
#define abcueindexsizev5 16
//*************************************************************************************
//Cue structure, version 5:
//The cue structure is varaible in size and padded with zeros up to an even number 
//of disk blocks (512 bytes). The cue structure contains all of the event structures in 
//the cue. If the cue is not in a real-time list, a cue time of zero signifies a manual 
//cue. A non-zero cue time specifies the time in milliseconds when the cue is to be 
//triggered by the list clock. If the cue is in a real-time list, the cue time is contained 
//in a real-time structure for that cue in the path structure; the cuetime longword 
//is set to zero. Bit 0 of the 'flags' word is the zero clock flag (list clock is zeroed 
//when the cue is run). All other bits are reserved, and should be set to zero.
struct abcuev5 {
	long size;							//size of cue structure, multiple of 512
	long unused1;					//used by AB64, zero this field
	long structID;					//constant 'ACUE' (0x41435545)
	long unused2;					//reserved, zero this field
	long cuenumber;				//major cue number (1..65535)
	unsigned char cueSub1;		//first sub number (0..255)
	unsigned char cueSub2;		//second sub number (0..255)
	unsigned char cueSub3;		//third sub number (0..255)
	unsigned char cueSub4;		//fourth sub number (0..255)
	long cuetime;					//cue time
	long flags;						//cue flags: bit 0: Zero Clock flag, all other bits reserved, set to zero
	long nevents;					//number of events in this cue
	char unused3[28];				//reserved, zero this field
	};
typedef struct abcuev5* abcuev5Ptr;
#define abcuev5size 64
//Variable size event heap immediately follows abcuev5 structure
//*************************************************************************************
//Event structure, version 5:
//The version 5 event structure is simply a single MIDI message prefixed with a length 
//and a delay value. There are no alignment restrictions; event structures are concatenated 
//without padding. The length field contains the total number of bytes in the MIDI 
//command (it does not include the first four bytes of the structure: the length and 
//delay bytes). The MIDI command must be 128 bytes or less in length. The event delay 
//is the interval in milliseconds from when the cue GOes to when the MIDI command is 
//sent to the MIDI OUT port and (optionally) to the AudioBox command queue. The 
//event delay is a 24-bit unsigned value stored in three consequtive bytes, most 
//significant byte first, 8 bits per byte.
struct abeventv5 {
	unsigned char MIDIlength;	//number of bytes in the MIDI string
	unsigned char evDelay1;		//event delay, msb
	unsigned char evDelay2;		//event delay, middle byte
	unsigned char evDelay3;		//event delay, lsb
	unsigned char MIDIcmd[128];	//variable length MIDI command string
	};
typedef struct abeventv5* abeventv5Ptr;
#define abeventv5HeaderSize 4
//*************************************************************************************
// End Version 5 Show File Structures
//*************************************************************************************
#define showEntryCount 128
#define audioEntryCountExt 8064
struct abdirectoryext {
	long diskCapacity;     //total capacity of the disk in blocks
	long nowfreeBlks;      //number of blocks now free on the disk
	long freeBlocks;       //number of blocks free after defragmentation
	long highestSelNum;       //highest selection number in use
	long showEntries[showEntryCount]; //array of show file entries
	long audioEntries[audioEntryCountExt];//array of audio selection entries
 	char audioChannels[audioEntryCountExt];//array of audio selection channel counts
};
//*************************************************************************************
struct diskStateStruct {
	long cmdID;
	char diskEnable;    //playback system is enabled; ready for playback
	char playbackActive;//one or more channels are playing back
	char dirError;      //error occured on last directory verify; one or more files were erased
	char transferActive;//audio or show file transfer is in progress
	char eraseActive;   //audio or show file erasure is in progress
	char defragActive;  //disk defragment is in progress
	char formatActive;  //disk format operation is in progress
	char dirClrActive;  //clear directory is in progress (CLEAR DATA)
	char MotorMixActive;//MotorMix interface has been activated
	char AllOffLock;    //ALLOFF lock
	char transferRate;   //reserved
	char audioMode;		//audio mode
	long diskCapacity;     //total capacity of the disk in blocks
	long nowfreeBlks;      //number of blocks now free on the disk
	long freeBlocks;       //number of blocks free after defragmentation
	long highestSelNum;       //highest selection number in use
	long defragMarker;  //defragment marker (for progress indicator)
	long defragEnd;     //defragment endpoint (for progress indicator)
	long highestShNum;       //highest show number in use
	char GPOstate;			//state of GPO outputs, bit 0: GPO 1, bit 1: GPO 2
	char autoConfIP;     //using auto-configured IP address
	char slaveDisk;      //0: using master/primary disk, 1: using slave/secondary 
	char unused2[25];   //reserved
	char MIDIID;        //MIDI device ID
	char MIDIchannel;   //assigned MIDI channel (FFH: all channels)
	short MIDIgroupIDs; //MIDI group IDs assigned
	char defaultShow;   //default show number
	char allOffActions; //MSC ALL_OFF action flags
	char resetActions1; //MSC RESET action flags, byte 1
	char resetActions2; //MSC RESET action flags, byte 2
	char submasterGain; //current submaster gain setting
	char submasterZero; //current submaster zero setting
	char fixedIP;       //fixed IP flag - set if ABFIXEDIP is in use (returned in INQUIRY)
	char GPIplbkTrigMode; //GPI playback trigger mode
	long GPItrigLockout; //GPI trigger lockout time in milliseconds
	char unused4[32];    //reserved
	};
//*************************************************************************************
struct cobraNetStruct {
	unsigned char code;
	char unused;
	short length;
	short CN1RXBundle[4];   //Module 1 receiver 0..3 bundle assignments
	short CN1TXBundle[4];   //Module 1 transmitter 0..3 bundle assignments
	short CN2RXBundle[4];   //Module 2 receiver 0..3 bundle assignments
	short CN2TXBundle[4];   //Module 2 transmitter 0..3 bundle assignments
	char reserved[256];       //reserved for future use
	};
typedef struct cobraNetStruct* cobranetPtr;
//*************************************************************************************
// filter types
enum { eFTypeNone = 0,  eFTypeLowShelf, eFTypeHighShelf, eFTypePeaking,
			eFTypeLowPass, eFTypeHighPass, eFTypeNotch, eFTypeBandpass};

//file types
enum { ftAudio = 0,  ftShow};
/*************************************************************************************
	function prototypes -- library call interface
*************************************************************************************/
int findAllAB64s (void);
struct ab64Inquiry *nextAB64Reply (void);
void connectAB64 (long IPaddress, long updateflags);
ab64MsgPtr nextAB64Message (void);
void sendCommand (unsigned char *cmd, long cmdLen);
void allAB64Data (void);
void setFileData (char *fileName, char *sourcePathName, long fileNumber, long fileType, long modTime, long transferTime);
int transferFileToAB64 (char *filePathName, char *sourcePathName, long fileNumber, long fileType);
int transferFileFromAB64 (char *filePath, long fileNumber, long fileType, long channels, long blocks);
void cancelTransfer (void);
void eraseAB64File (long fileNumber, long fileType);
int transferFirmware (char *filePath, int writeToDisk);
void hostTov4ShowFile (void *shPtr, int shoSize);
void hostTov5ShowFile (void *shPtr, int shoSize);
void v4showFileToHost (void *shPtr, int shoSize);
void v5showFileToHost (void *shPtr, int shoSize);
void roundUpToPage (long *x);

// AudioBox commands
void allOff (void); //show ALL OFF
void allOnMatrix (void); //sets to full on all inputs, outputs and crosspoints
void assignInputsToSubMaster (char *assignString, long submaster); //assigns input channels to a submaster. assign string is comma delimited, can contain ranges (i.e. "1-4, 10, 12-15")
void assignOutputsToSubMaster (char *assignString, long submaster); //assigns output channels to a submaster. assign string is comma delimited, can contain ranges (i.e. "1-4, 10, 12-15")
void assignSubMasterToController (long submaster, long controller); //assigns a submaster to a MIDI controller
void clearAllDelays (void); //clears all input, output and crosspoint delays
void clearAssigns (void); //clears all submasters and controller assigns (all submasters, all controllers)
void clearDirectory (void);  //clears the directory on the AudioBox
void clearFirmware (void); //test purposes only
void clearSettings (void); //clears all settings: MIDI ID, submasters, gain tables, etc.
void clearMessageCounts (void); //zeros all message counts and  low water marks returned by GET STATS
void clearSubmasterAssigns (long submaster, long IO); //clears all input (IO=0) or output (IO=1) channel assignments from a submaster
void closeList (long list); //closes list
void closePath (long list, long path); //closes path
void deassignSubMasterFromController (long submaster, long controller); //deassigns a submaster from a MIDI controller
void decrementStandbyCue (long list); //decrement standby cue
void diagonalMatrix (void); //sets full on all inputs, outpus and diagonal crosspoints (1:1, 2:2, ...)
void disablePlaybackChannel (long ch); //channel numbers 'ch' are 1..64
void pushFrontPanelButton (long button); //remotely pushed front panel button 0:left, 1:middle, 2:right
void enableRecordChannel (long channel, long enable); //enables channel for recording, enable = 1 to enable channel, 0 to disable channel
void FrequencyResponse (double theta, double inA2, double inA1, double inA0, double inB1, double inB0, double *outMag); //provide a point on an EQ curve
void goAllPlayback (void); //sends a null GO, starting all enabled playback channels
void goCue (long cue, long sub1, long sub2, long sub3, long sub4, long list, long path); //fires a cue
void goCueJamClock (long cue, long sub1, long sub2, long sub3, long sub4, long list, long path); //fires cue and sets list clock to cue time
void goPlayback (long ch, long sel, long min, long sec, long frame, long framesPerSecond); // channel numbers 'ch' are 1..64
void goPlaybackChannels (long chSet1, long chSet2); //ch1=chSet1:bit0, ch33=chSet2:bit0, ch64=chSet2:bit31
void incrementStandbyCue (long list); //increment standby cue
void loadCue (long cue, long sub1, long sub2, long sub3, long sub4, long list, long path); //loads cue to standby
void loadPlaybackMultiChannel (long plbkch, long sel, long min, long sec, long frame, long framesPerSecond); 
void openList (long show, long list); //opens list, and show if show not already open
void openPath (long list, long path); //opens path, and list if list not already open
void playbackStop (long plbkch); // channel numbers 'ch' are 1..64
void reset (void); //show RESET
void resumeCue (long cue, long sub1, long sub2, long sub3, long sub4, long list);
void setAB64Name (char *name); //sets AB64 name string
void setAllOffActions (long flags); //sets ALL OFF action flags (see command set for bit definitions)
void setCobraNet (short *bundleArray);	//set CobraNet bundles. bundleArray is a pointer to an array of 16 shorts (see command set: SET COBRANET EN)
void setListClock (long hour, long minute, long second, long frame, long subframe, long list, long framesPerSecond); //set list clock
long setConnectedAB (long index); //sets ab64if to one of the connected AudioBoxes, returns false if error occurred
void setCrosspointDelay (long inch, long ouch, long dly, long hour, long minute, long second, long frame, long subframe, long framesPerSecond); //set crosspoint delay
void setCrosspointLevel (long inch, long ouch, long level, long exp, long hour, long minute, long second, long frame, long subframe, long framesPerSecond); //channel numbers 'inch' are 1..64
void setCrosspointPhase (long inch, long ouch, long phase); //channel numbers 'inch' are 1..64
void setDefaultShow (long show); //sets the default show (writes settings to disk)
void setGPI (long mode, long lockout); //sets GPI parameters
void setGPO (long output, long state);  //sets GPO state
void setInputDelay (long ch, long dly); //dly in samples, channel numbers 'ch' are 1..64
void setInputDelayEnable (long ch, long enable); //enable: 1, disable: 0, channel numbers 'ch' are 1..64
long setInputEQ (long ch, long band, long freq, long bw, long fg, long type); // channel numbers 'ch' are 1..64, returns false if filter cannot be realized
void setInputEQbands (long ch, long bands); //sets number of EQ bands on input channel, channel numbers 'ch' are 1..64
void setInputLevel (long ch, long level, long exp, long hour, long minute, long second, long frame, long subframe, long framesPerSecond); //channel numbers 'ch' are 1..64
void setInputMute (long ch, long mute); // mute: 1, unmute: 0, channel numbers 'ch' are 1..64
void setInputSolo (long ch, long solo); // solo: 1, unsolo: 0, channel numbers 'ch' are 1..64
void setMidiId (long id); //set MIDI ID (writes settings to disk)
void setMidiEcho (long midi, long show, long stopNotify);	//set MIDI echo flags
void setMTCchaseOn (long list);	//turn MTC chase on
void setMTCchaseOff (long list);	//turn MTC chase off
void setMTCsourceInternal (long list);	//make internal MTC generator source of MTC for list
void setMTCsourceExternal (long list);	//make external MTC source of MTC for list
void setOutputDelay (long ch, long dly); //dly in samples, channel numbers 'ch' are 1..64
void setOutputDelayEnable (long ch, long enable); //enable: 1, disable: 0, channel numbers 'ch' are 1..64
long setOutputEQ (long ch, long band, long freq, long bw, long fg, long type); // channel numbers 'ch' are 1..64, returns false if filter cannot be realized
void setOutputEQbands (long ch, long bands);//sets number of EQ bands on output channel, channel numbers 'ch' are 1..64
void setOutputLevel (long ch, long level, long exp, long hour, long minute, long second, long frame, long subframe, long framesPerSecond); //channel numbers 'ch' are 1..64
void setOutputMute (long ch, long mute); // mute: 1, unmute: 0, channel numbers 'ch' are 1..64
void setOutputSolo (long ch, long solo); // solo: 1, unsolo: 0, channel numbers 'ch' are 1..64
void setPlaybackLocation (long ch, long sel, long min, long sec, long frame, long framesPerSecond); // channel numbers 'ch' are 1..64
void setPlaybackBlockLocation (long ch, long sel, long block);
void setPlaybackResume (long ch, long sel, long min, long sec, long frame, long framesPerSecond); // channel numbers 'ch' are 1..64
void setPlaybackStop (long ch, long sel, long min, long sec, long frame, long framesPerSecond); // channel numbers 'ch' are 1..64
void setRealTimeClock (void);	//set AudioBox real-time clock
void setResetActions (long flag1, long flag2); //sets RESET action flags (see command set for bit definitions)
void setResumeBlockLocation (long ch, long sel, long block);
void setSourceMix (long ch, long analog, long digital, long playback); //sets mix of input signals (on AB1616 / 1616HD sets input channel mode)
void setStopBlockLocation (long ch, long sel, long block);
void setSubMasterGain (long gain); //sets submaster gain. 0..3
void setSubMasterLevel (long submaster, long level); //sets submaster level (submaster 1..32, MIDI level 0..127)
void setSubMasterZero (long zero); //sets submaster zero (MIDI level 0..127)
void setupRecord (long selection, long inOut); // inOut ==  0: record inputs, 1: record outputs
void startListClock (long list); //start list clock
void stopAllPlayback (void); //stops all playback
void stopListClock (long list); //stop list clock
void stopCue (long cue, long sub1, long sub2, long sub3, long sub4, long list);
void turnOffMatrix (void); //sets to zero all inputs, outputs and crosspoints
void zeroListClock (long list); //zero list clock















